#    Copyright 2020, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import copy
import imp

try:
    from unittest import mock
except ImportError:
    import mock
from oslo_config import cfg
from oslo_config import fixture as oslo_fixture
from oslo_utils import uuidutils

from octavia.common import data_models as o_data_models
from octavia.common import exceptions as o_exceptions
from octavia.network import data_models as n_data_models
from octavia.tests.common import constants as t_constants

from a10_octavia.common import config_options
from a10_octavia.common import data_models
from a10_octavia.common import exceptions
from a10_octavia.common import utils
from a10_octavia.controller.worker.tasks import a10_database_tasks as task
from a10_octavia.tests.common import a10constants
from a10_octavia.tests.unit import base

VTHUNDER = data_models.VThunder(amphora_id=a10constants.MOCK_AMPHORA_ID,
                                compute_id=a10constants.MOCK_COMPUTE_ID)
HW_THUNDER = data_models.HardwareThunder(
    project_id=a10constants.MOCK_PROJECT_ID,
    device_name="rack_thunder_1",
    undercloud=True,
    username="abc",
    password="abc",
    ip_address="10.10.10.10",
    partition_name="shared")
LB = o_data_models.LoadBalancer(id=a10constants.MOCK_LOAD_BALANCER_ID,
                                flavor_id=a10constants.MOCK_FLAVOR_ID)
LB2 = o_data_models.LoadBalancer(id=a10constants.MOCK_LOAD_BALANCER_ID)
FIXED_IP = n_data_models.FixedIP(ip_address='10.10.10.10')
PORT = n_data_models.Port(id=uuidutils.generate_uuid(), fixed_ips=[FIXED_IP])
VRID = data_models.VRID(id=1, vrid=0,
                        owner=a10constants.MOCK_PROJECT_ID,
                        vrid_port_id=uuidutils.generate_uuid(),
                        vrid_floating_ip='10.0.12.32')
VRRP_SET = data_models.VrrpSet(set_id=a10constants.MOCK_SET_ID)
LISTENER = o_data_models.Listener(id=a10constants.MOCK_LISTENER_ID, load_balancer=LB)
POOL = o_data_models.Pool(id=a10constants.MOCK_POOL_ID, load_balancer=LB)
HM = o_data_models.HealthMonitor(id=a10constants, pool=POOL)
MEMBER_1 = o_data_models.Member(id=uuidutils.generate_uuid(),
                                project_id=a10constants.MOCK_PROJECT_ID,
                                subnet_id=a10constants.MOCK_SUBNET_ID,
                                pool=POOL)
MEMBER_2 = o_data_models.Member(id=uuidutils.generate_uuid(),
                                project_id=a10constants.MOCK_PROJECT_ID,
                                subnet_id=a10constants.MOCK_SUBNET_ID_2,
                                pool=POOL)
FLAVOR_PROFILE = o_data_models.FlavorProfile(id=a10constants.MOCK_FLAVOR_PROF_ID,
                                             flavor_data={})
FLAVOR = o_data_models.Flavor(id=a10constants.MOCK_FLAVOR_ID,
                              flavor_profile_id=a10constants.MOCK_FLAVOR_PROF_ID)
SUBNET = n_data_models.Subnet(id=uuidutils.generate_uuid())
NAT_POOL = data_models.NATPool(id=uuidutils.generate_uuid(),
                               port_id=a10constants.MOCK_PORT_ID)

AMPHORA = o_data_models.Amphora(id=a10constants.MOCK_AMPHORA_ID,
                                role=a10constants.MOCK_TOPOLOGY_ROLE_MASTER)
AMPHORA_BLADE = o_data_models.Amphora(id=a10constants.MOCK_AMPHORA_ID,
                                      role=a10constants.MOCK_TOPOLOGY_ROLE_BACKUP)


class TestA10DatabaseTasks(base.BaseTaskTestCase):

    def setUp(self):
        super(TestA10DatabaseTasks, self).setUp()
        imp.reload(task)
        self.vrid_repo = mock.Mock()
        self.nat_pool_repo = mock.Mock()
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        self.conf.register_opts(config_options.A10_GLOBAL_OPTS,
                                group=a10constants.A10_GLOBAL_CONF_SECTION)
        self.conf.register_opts(
            config_options.A10_HARDWARE_THUNDER_OPTS,
            group=a10constants.HARDWARE_THUNDER_CONF_SECTION)
        self.conf.register_opts(config_options.A10_GLOBAL_OPTS,
                                group=a10constants.A10_GLOBAL_OPTS)
        self.db_session = mock.patch(
            'a10_octavia.controller.worker.tasks.a10_database_tasks.db_apis.get_session')
        self.db_session.start()

    def tearDown(self):
        super(TestA10DatabaseTasks, self).tearDown()
        self.db_session.stop()

    def _generate_hardware_device_conf(self, thunder):
        hardware_device_conf = [{"project_id": a10constants.MOCK_PROJECT_ID,
                                 "username": thunder.username,
                                 "password": thunder.password,
                                 "device_name": thunder.device_name,
                                 "vrid_floating_ip": thunder.vrid_floating_ip,
                                 "ip_address": thunder.ip_address}]
        return hardware_device_conf

    def test_get_vthunder_by_loadbalancer(self):
        self.conf.config(
            group=a10constants.A10_GLOBAL_CONF_SECTION,
            use_parent_partition=True)
        mock_get_vthunder = task.GetVThunderByLoadBalancer()
        mock_get_vthunder.vthunder_repo = mock.MagicMock()
        mock_get_vthunder.vthunder_repo.get_vthunder_from_lb.return_value = None
        vthunder = mock_get_vthunder.execute(LB)
        self.assertEqual(vthunder, None)

    def test_get_backup_vthunder_by_loadbalancer(self):
        vthunder = copy.deepcopy(VTHUNDER)
        vthunder.role = "BACKUP"
        self.conf.config(
            group=a10constants.A10_GLOBAL_CONF_SECTION,
            use_parent_partition=True)
        mock_get_vthunder = task.GetVThunderByLoadBalancer()
        mock_get_vthunder.vthunder_repo = mock.MagicMock()
        mock_get_vthunder.vthunder_repo.get_vthunder_from_lb.return_value = None
        mock_get_vthunder.vthunder_repo.get_backup_vthunder_from_lb.return_value = vthunder
        backup_vthunder = mock_get_vthunder.execute(LB, False)
        self.assertEqual(backup_vthunder, vthunder)

    def test_get_vrid_for_project_member_hmt_use_partition(self):
        thunder = copy.deepcopy(HW_THUNDER)
        lb = copy.deepcopy(LB)
        thunder.hierarchical_multitenancy = 'enable'
        thunder.vrid_floating_ip = VRID.vrid_floating_ip
        hardware_device_conf = self._generate_hardware_device_conf(thunder)
        self.conf.config(group=a10constants.HARDWARE_THUNDER_CONF_SECTION,
                         devices=[hardware_device_conf])
        self.conf.config(
            group=a10constants.A10_GLOBAL_CONF_SECTION,
            use_parent_partition=True)
        self.conf.conf.hardware_thunder.devices = {
            a10constants.MOCK_PROJECT_ID: thunder}

        mock_vrid_entry = task.GetVRIDForLoadbalancerResource()
        mock_vrid_entry.CONF = self.conf

        mock_vrid_entry.vrid_repo = mock.Mock()
        mock_vrid_entry.vthunder_repo = mock.Mock()
        mock_vrid_entry.vrid_repo.get_vrid_from_owner.return_value = VRID
        vrid = mock_vrid_entry.execute([a10constants.MOCK_PROJECT_ID], thunder, lb)
        self.assertEqual(VRID, vrid)

    def test_get_vrid_for_project(self):
        thunder = copy.deepcopy(HW_THUNDER)
        lb = copy.deepcopy(LB)
        thunder.hierarchical_multitenancy = 'disable'
        thunder.vrid_floating_ip = VRID.vrid_floating_ip
        hardware_device_conf = self._generate_hardware_device_conf(thunder)
        self.conf.config(group=a10constants.HARDWARE_THUNDER_CONF_SECTION,
                         devices=[hardware_device_conf])
        self.conf.config(
            group=a10constants.A10_GLOBAL_CONF_SECTION,
            use_parent_partition=False)
        self.conf.conf.hardware_thunder.devices = {
            a10constants.MOCK_PROJECT_ID: thunder}

        mock_vrid_entry = task.GetVRIDForLoadbalancerResource()
        mock_vrid_entry.CONF = self.conf

        mock_vrid_entry.vrid_repo = mock.Mock()
        mock_vrid_entry.vthunder_repo = mock.Mock()
        mock_vrid_entry.vrid_repo.get_vrid_from_owner.return_value = VRID
        vrid = mock_vrid_entry.execute(MEMBER_1, thunder, lb)
        mock_vrid_entry.vthunder_repo.get_partition_for_project.assert_not_called()
        self.assertEqual(VRID, vrid)

    def test_update_vrid_for_project_member_delete_vrid(self):
        thunder = copy.deepcopy(HW_THUNDER)
        thunder.vrid_floating_ip = VRID.vrid_floating_ip
        hardware_device_conf = self._generate_hardware_device_conf(thunder)
        self.conf.config(group=a10constants.HARDWARE_THUNDER_CONF_SECTION,
                         devices=[hardware_device_conf])
        self.conf.conf.hardware_thunder.devices = {
            a10constants.MOCK_PROJECT_ID: thunder}
        utils.get_vrid_floating_ip_for_project.CONF = self.conf

        mock_vrid_entry = task.UpdateVRIDForLoadbalancerResource()
        mock_vrid_entry.vrid_repo = mock.Mock()
        mock_vrid_entry.execute(MEMBER_1, [], thunder)
        mock_vrid_entry.vrid_repo.delete.assert_called_once_with(
            mock.ANY,
            owner=thunder.ip_address + "_" + thunder.partition_name
        )

    def test_update_vrid_for_project_member_with_port_and_with_vrid(self):
        thunder = copy.deepcopy(HW_THUNDER)
        mock_vrid_entry = task.UpdateVRIDForLoadbalancerResource()
        mock_vrid_entry.vrid_repo = mock.Mock()
        mock_vrid_entry.vrid_repo.exists.return_value = True
        mock_vrid_entry.execute(MEMBER_1, [VRID], thunder)
        mock_vrid_entry.vrid_repo.update.assert_called_once_with(
            mock.ANY,
            VRID.id,
            vrid=VRID.vrid,
            vrid_floating_ip=VRID.vrid_floating_ip,
            vrid_port_id=VRID.vrid_port_id,
            subnet_id=VRID.subnet_id)
        mock_vrid_entry.vrid_repo.create.assert_not_called()

    def test_update_vrid_for_project_member_with_port_and_no_vrid(self):
        thunder = copy.deepcopy(HW_THUNDER)
        mock_vrid_entry = task.UpdateVRIDForLoadbalancerResource()
        mock_vrid_entry.vrid_repo = mock.Mock()
        mock_vrid_entry.vrid_repo.exists.return_value = False
        mock_vrid_entry.execute(MEMBER_1, [VRID], thunder)
        mock_vrid_entry.vrid_repo.create.assert_called_once_with(
            mock.ANY,
            owner=thunder.ip_address + "_" + thunder.partition_name,
            id=VRID.id,
            vrid=VRID.vrid,
            vrid_floating_ip=VRID.vrid_floating_ip,
            vrid_port_id=VRID.vrid_port_id,
            subnet_id=VRID.subnet_id)
        mock_vrid_entry.vrid_repo.update.assert_not_called()

    def test_delete_vrid_entry_with_vrid(self):
        mock_vrid_entry = task.DeleteVRIDEntry()
        mock_vrid_entry.vrid_repo = mock.Mock()
        mock_vrid_entry.execute(VRID, True)
        mock_vrid_entry.vrid_repo.delete.assert_called_once_with(
            mock.ANY, id=VRID.id)

    def test_delete_vrid_entry_negative(self):
        mock_vrid_entry = task.DeleteVRIDEntry()
        mock_vrid_entry.vrid_repo = mock.Mock()
        mock_vrid_entry.execute(VRID, False)
        mock_vrid_entry.vrid_repo.delete.assert_not_called()
        mock_vrid_entry.vrid_repo.reset_mock()
        mock_vrid_entry.execute(None, True)
        mock_vrid_entry.vrid_repo.delete.assert_not_called()
        mock_vrid_entry.vrid_repo.reset_mock()
        mock_vrid_entry.execute(None, False)
        mock_vrid_entry.vrid_repo.delete.assert_not_called()

    def test_delete_vrid_entry_with_multi_vrids(self):
        VRID_1 = data_models.VRID(id=2)
        mock_vrid_entry = task.DeleteMultiVRIDEntry()
        mock_vrid_entry.vrid_repo = mock.Mock()
        mock_vrid_entry.execute([VRID, VRID_1])
        mock_vrid_entry.vrid_repo.delete_batch.assert_called_once_with(
            mock.ANY, ids=[VRID.id, VRID_1.id])

    def test_count_members_in_project_ip(self):
        mock_count_member = task.CountMembersWithIP()
        mock_count_member.member_repo.get_member_count_by_ip_address = mock.Mock()
        mock_count_member.member_repo.get_member_count_by_ip_address.return_value = 1
        member_count = mock_count_member.execute(MEMBER_1)
        self.assertEqual(1, member_count)

    def test_count_members_in_project_ip_port_protocol(self):
        member_1 = o_data_models.Member(
            id=uuidutils.generate_uuid(),
            protocol_port=t_constants.MOCK_PORT_ID,
            project_id=t_constants.MOCK_PROJECT_ID,
            ip_address=t_constants.MOCK_IP_ADDRESS)
        mock_count_member = task.CountMembersWithIPPortProtocol()
        mock_count_member.member_repo.get_member_count_by_ip_address_port_protocol = mock.Mock()
        mock_count_member.member_repo.get_member_count_by_ip_address_port_protocol.return_value = 2
        member_count = mock_count_member.execute(member_1, POOL)
        self.assertEqual(2, member_count)

    def test_pool_count_accn_ip(self):
        use_dev_flavor = False
        member_1 = o_data_models.Member(id=uuidutils.generate_uuid(),
                                        project_id=t_constants.MOCK_PROJECT_ID,
                                        ip_address=t_constants.MOCK_IP_ADDRESS,
                                        pool_id=a10constants.MOCK_POOL_ID)
        mock_count_pool = task.PoolCountforIP()
        mock_count_pool.member_repo.get_pool_count_by_ip = mock.Mock()
        mock_count_pool.member_repo.get_pool_count_by_ip.return_value = 2
        pool_count = mock_count_pool.execute(member_1, use_dev_flavor, mock.ANY)
        self.assertEqual(2, pool_count)

    def test_get_subnet_for_deletion_with_pool_no_lb(self):
        use_dev_flavor = False
        mock_subnets = task.GetSubnetForDeletionInPool()
        mock_subnets.member_repo.get_pool_count_subnet = mock.Mock()
        mock_subnets.member_repo.get_pool_count_subnet.return_value = 1
        mock_subnets.loadbalancer_repo.get_lb_count_by_subnet = mock.Mock()
        mock_subnets.loadbalancer_repo.get_lb_count_by_subnet.return_value = 0
        subnet_list = mock_subnets.execute(
            [MEMBER_1, MEMBER_2], [a10constants.MOCK_PROJECT_ID], use_dev_flavor, mock.ANY)
        self.assertEqual(['mock-subnet-1', 'mock-subnet-2'], subnet_list)

    def test_get_subnet_for_deletion_with_multiple_pool_lb(self):
        use_dev_flavor = False
        mock_subnets = task.GetSubnetForDeletionInPool()
        mock_subnets.member_repo.get_pool_count_subnet = mock.Mock()
        mock_subnets.member_repo.get_pool_count_subnet.return_value = 2
        mock_subnets.loadbalancer_repo.get_lb_count_by_subnet = mock.Mock()
        mock_subnets.loadbalancer_repo.get_lb_count_by_subnet.return_value = 2
        subnet_list = mock_subnets.execute(
            [MEMBER_1, MEMBER_2], [a10constants.MOCK_PROJECT_ID], use_dev_flavor, mock.ANY)
        self.assertEqual([], subnet_list)

    def test_get_child_projects_for_partition(self):
        vthunder = copy.deepcopy(VTHUNDER)
        vthunder.partition_name = "mock-partition-name"
        vthunder.ip_address = "mock-ip-addr"
        mock_get_projects = task.GetChildProjectsOfParentPartition()
        mock_get_projects.vthunder_repo.get_project_list_using_partition = mock.Mock()
        mock_get_projects.execute(MEMBER_1, vthunder)
        mock_get_projects.vthunder_repo.get_project_list_using_partition.\
            assert_called_once_with(mock.ANY, partition_name='mock-partition-name',
                                    ip_address="mock-ip-addr")

    def test_GetFlavorData_format_flavor_keys(self):
        expected = {"virtual_server": {"arp_disable": 1}}
        flavor = {"virtual-server": {"arp-disable": 1}}
        flavor_task = task.GetFlavorData()
        formated_flavor = flavor_task._format_keys(flavor)
        self.assertEqual(formated_flavor, expected)

    def test_GetFlavorData_format_flavor_list_keys(self):
        expected = {
            "service_group": {
                "name_expressions": [{
                    "regex": "sg1",
                    "json": {"health_check_disable": 1}
                }]
            }
        }
        name_expr = {
            "name-expressions": [{
                "regex": "sg1",
                "json": {"health-check-disable": 1}
            }]
        }
        flavor = {"service-group": name_expr}
        flavor_task = task.GetFlavorData()
        formated_flavor = flavor_task._format_keys(flavor)
        self.assertEqual(formated_flavor, expected)

    def test_GetFlavorData_format_flavor_multi_val(self):
        expected = {
            "service_group": {
                "name_expressions": [{
                    "regex": "sg1",
                    "json": {"health_check_disable": 1}
                }],
                "strict_select": 0
            }
        }
        name_expr = {
            "name-expressions": [{
                "regex": "sg1",
                "json": {"health-check-disable": 1}
            }]
        }
        flavor = {"service-group": {"strict-select": 0}}
        flavor['service-group'].update(name_expr)
        flavor_task = task.GetFlavorData()
        formated_flavor = flavor_task._format_keys(flavor)
        self.assertEqual(formated_flavor, expected)

    @mock.patch('a10_octavia.controller.worker.tasks.a10_database_tasks.a10_task_utils')
    def test_GetFlavorData_execute_no_flavor_id(self, mock_utils):
        mock_utils.attribute_search.return_value = None
        flavor_task = task.GetFlavorData()
        ret_val = flavor_task.execute(LB)
        self.assertEqual(ret_val, None)

    def test_GetFlavorData_execute_no_flavor_or_profile(self):
        flavor_task = task.GetFlavorData()
        flavor_task.flavor_repo = mock.Mock()
        flavor_task.flavor_repo.get.return_value = None
        flavor_task._flavor_search = mock.Mock(
            return_value=a10constants.MOCK_FLAVOR_ID)
        ret_val = flavor_task.execute
        self.assertRaises(exceptions.FlavorNotFound, ret_val, LB)

        flavor_task.flavor_repo.reset_mock()
        flavor_task.flavor_repo = mock.Mock()
        flavor = copy.deepcopy(FLAVOR)
        flavor.flavor_profile_id = None
        flavor_task.flavor_repo.get.return_value = flavor
        ret_val = flavor_task.execute(LB)
        self.assertEqual(ret_val, None)

    def test_GetFlavorData_execute_with_default_flavor_id(self):
        flavor_task = task.GetFlavorData()
        flavor = copy.deepcopy(FLAVOR)
        flavor_prof = copy.deepcopy(FLAVOR_PROFILE)
        flavor_prof.flavor_data = "{}"
        flavor_task.flavor_repo = mock.Mock()
        flavor_task.flavor_repo.get.return_value = flavor
        self.conf.config(group=a10constants.A10_GLOBAL_OPTS,
                         default_flavor_id=flavor)
        flavor_task.flavor_profile_repo = mock.Mock()
        flavor_task.flavor_profile_repo.get.return_value = flavor_prof
        ret_val = flavor_task.execute(LB2)
        self.assertEqual(ret_val, {})

    def test_GetFlavorData_execute_return_flavor(self):
        flavor_task = task.GetFlavorData()
        flavor_task._flavor_search = mock.Mock(
            return_value=a10constants.MOCK_FLAVOR_ID)
        flavor_task._format_keys = mock.Mock()
        flavor_task._format_keys.return_value = {}
        flavor_task.flavor_repo = mock.Mock()
        flavor_task.flavor_repo.get.return_value = FLAVOR

        flavor_prof = copy.deepcopy(FLAVOR_PROFILE)
        flavor_prof.flavor_data = "{}"
        flavor_task.flavor_profile_repo = mock.Mock()
        flavor_task.flavor_profile_repo.get.return_value = flavor_prof
        ret_val = flavor_task.execute(LB)
        self.assertEqual(ret_val, {})

    def test_GetNatPoolEntry(self):
        db_task = task.GetNatPoolEntry()
        db_task.nat_pool_repo = self.nat_pool_repo
        flavor = {"pool_name": "p1", "start_address": "1.1.1.1", "end_address": "1.1.1.2"}
        db_task.execute(MEMBER_1, flavor)
        self.nat_pool_repo.get(mock.ANY, name="p1", subnet_id=MEMBER_1.subnet_id)

    def test_UpdateNatPoolDB_update(self):
        db_task = task.UpdateNatPoolDB()
        db_task.nat_pool_repo = self.nat_pool_repo
        flavor = {"pool_name": "p1", "start_address": "1.1.1.1", "end_address": "1.1.1.2"}
        NAT_POOL.member_ref_count = 1
        db_task.execute(MEMBER_1, flavor, NAT_POOL, None)
        self.nat_pool_repo.update.assert_called_with(mock.ANY, mock.ANY, member_ref_count=2)

    def test_UpdateNatPoolDB_create(self):
        db_task = task.UpdateNatPoolDB()
        db_task.nat_pool_repo = self.nat_pool_repo
        flavor = {"pool_name": "p1", "start_address": "1.1.1.1", "end_address": "1.1.1.2"}
        port = mock.Mock(id="1")
        db_task.execute(MEMBER_1, flavor, None, port)
        self.nat_pool_repo.create.assert_called_with(
            mock.ANY, id=mock.ANY, name="p1",
            subnet_id=MEMBER_1.subnet_id, start_address="1.1.1.1", end_address="1.1.1.2",
            member_ref_count=1, port_id="1")

    def test_DeleteNatPoolEntry_update(self):
        db_task = task.DeleteNatPoolEntry()
        db_task.nat_pool_repo = self.nat_pool_repo
        NAT_POOL.member_ref_count = 2
        db_task.execute(NAT_POOL)
        self.nat_pool_repo.update.assert_called_with(mock.ANY, mock.ANY, member_ref_count=1)

    def test_DeleteNatPoolEntry_delete(self):
        db_task = task.DeleteNatPoolEntry()
        db_task.nat_pool_repo = self.nat_pool_repo
        NAT_POOL.member_ref_count = 1
        db_task.execute(NAT_POOL)
        self.nat_pool_repo.delete(mock.ANY, id=NAT_POOL.id)

    def test_lb_count_according_to_flavor(self):
        mock_count_lb = task.CountLoadbalancersWithFlavor()
        mock_count_lb.loadbalancer_repo.get_lb_count_by_flavor = mock.Mock()
        mock_count_lb.loadbalancer_repo.get_lb_count_by_flavor.return_value = 2
        lb_count = mock_count_lb.execute(LB, VTHUNDER)
        self.assertEqual(2, lb_count)

    def test_SetThunderUpdatedAt_execute_update(self):
        db_task = task.SetThunderUpdatedAt()
        vthunder = copy.deepcopy(VTHUNDER)
        db_task.vthunder_repo.update = mock.Mock()
        db_task.execute(vthunder)
        db_task.vthunder_repo.update.assert_called_once_with(mock.ANY,
                                                             vthunder.id,
                                                             updated_at=mock.ANY)

    def test_SetThunderLastWriteMem_execute_update(self):
        db_task = task.SetThunderLastWriteMem()
        vthunder = copy.deepcopy(VTHUNDER)
        db_task.vthunder_repo.update_last_write_mem = mock.Mock()
        db_task.execute(vthunder, True, True)
        db_task.vthunder_repo.update_last_write_mem.assert_called_once_with(
            mock.ANY, vthunder.ip_address, vthunder.partition_name, last_write_mem=mock.ANY)

    def test_GetActiveLoadBalancersByThunder_return_empty(self):
        lb_task = task.GetActiveLoadBalancersByThunder()
        vthunder = copy.deepcopy(VTHUNDER)
        vthunder.loadbalancer_id = a10constants.MOCK_LOAD_BALANCER_ID
        lb_task.loadbalancer_repo.get_active_lbs_by_thunder = mock.Mock()
        lb_task.loadbalancer_repo.get_active_lbs_by_thunder.return_value = []
        lb_list = lb_task.execute(vthunder)
        self.assertEqual(lb_list, [])

    def test_GetActiveLoadBalancersByThunder_return_list(self):
        lb_task = task.GetActiveLoadBalancersByThunder()
        vthunder = copy.deepcopy(VTHUNDER)
        vthunder.loadbalancer_id = a10constants.MOCK_LOAD_BALANCER_ID
        lb_task.loadbalancer_repo.get_active_lbs_by_thunder = mock.Mock()
        lb_task.loadbalancer_repo.get_active_lbs_by_thunder.return_value = [LB]
        lb_list = lb_task.execute(vthunder)
        self.assertEqual(len(lb_list), 1)

    def test_MarkLoadBalancersPendingUpdateInDB_execute(self):
        lb_task = task.MarkLoadBalancersPendingUpdateInDB()
        lb_list = []
        lb_list.append(LB)
        lb_task.loadbalancer_repo.update = mock.Mock()
        lb_task.execute(lb_list)
        lb_task.loadbalancer_repo.update.assert_called_once_with(mock.ANY,
                                                                 LB.id,
                                                                 provisioning_status=mock.ANY)

    def test_MarkLoadBalancersPendingUpdateInDB_execute_for_empty_list(self):
        lb_task = task.MarkLoadBalancersPendingUpdateInDB()
        lb_list = []
        lb_task.loadbalancer_repo.update = mock.Mock()
        lb_task.execute(lb_list)
        lb_task.loadbalancer_repo.update.assert_not_called()

    def test_MarkLoadBalancersActiveInDB_execute(self):
        lb_task = task.MarkLoadBalancersActiveInDB()
        lb_list = []
        lb_list.append(LB)
        lb_task.loadbalancer_repo.update = mock.Mock()
        lb_task.execute(lb_list)
        lb_task.loadbalancer_repo.update.assert_called_once_with(mock.ANY,
                                                                 LB.id,
                                                                 provisioning_status=mock.ANY)

    def test_MarkLoadBalancersActiveInDB_execute_for_empty_list(self):
        lb_task = task.MarkLoadBalancersActiveInDB()
        lb_list = []
        lb_task.loadbalancer_repo.update = mock.Mock()
        lb_task.execute(lb_list)
        lb_task.loadbalancer_repo.update.assert_not_called()

    def test_AddProjectSetIdDB(self):
        self.conf.config(
            group=a10constants.A10_CONTROLLER_WORKER_CONF_SECTION,
            amp_boot_network_list="mgmt_subnet_1")
        db_task = task.AddProjectSetIdDB()
        db_task.vrrp_set_repo.get = mock.Mock()
        db_task.execute(LB)
        db_task.vrrp_set_repo.get.assert_called_once_with(mock.ANY, mgmt_subnet='mgmt_subnet_1',
                                                          project_id=mock.ANY)

    def test_DeleteProjectSetIdDB(self):
        db_task = task.DeleteProjectSetIdDB()
        db_task.vrrp_set_repo.delete = mock.Mock()
        db_task.execute(LB)
        db_task.vrrp_set_repo.delete.assert_called_once_with(mock.ANY, project_id=mock.ANY)

    def test_CheckExistingVthunderTopology_execute_raised_exception(self):
        lb_task = task.CheckExistingVthunderTopology()
        vthunder = copy.deepcopy(VTHUNDER)
        vthunder.topology = a10constants.MOCK_TOPOLOGY_ACTIVE_STANDBY
        lb_task.vthunder_repo = mock.MagicMock()[1:999]
        lb_task.vthunder_repo.get_vthunder_by_project_id.return_value = vthunder
        self.assertRaises(o_exceptions.InvalidTopology, lb_task.execute, LB, "SINGLE")

    @mock.patch('octavia.statistics.stats_base.update_stats_via_driver')
    def test_update_listeners_stats_with_statistics(self, mock_stats_base):
        LISTENER_STATS = o_data_models.ListenerStatistics(listener_id=uuidutils.generate_uuid())
        mock_get_listener = task.UpdateListenersStats()
        mock_get_listener.listener_repo = mock.MagicMock()
        mock_get_listener.listener_repo.get_all.return_value = [LISTENER], None
        mock_get_listener.listener_stats_repo = mock.MagicMock()
        mock_get_listener.listener_stats_repo.get_all.return_value = [LISTENER_STATS], None
        mock_get_listener.execute([LISTENER_STATS])
        mock_stats_base.assert_called_once_with([LISTENER_STATS])

    def test_GetVThunderAmphora(self):
        db_task = task.GetVThunderAmphora()
        vthunder = copy.deepcopy(VTHUNDER)
        db_task.amphora_repo = mock.MagicMock()
        db_task.amphora_repo.get.return_value = AMPHORA
        amp = db_task.execute(vthunder)
        self.assertEqual(amp.id, a10constants.MOCK_AMPHORA_ID)

    def test_GetProjectVRRPSetId(self):
        db_task = task.GetProjectVRRPSetId()
        self.conf.config(
            group=a10constants.A10_CONTROLLER_WORKER_CONF_SECTION,
            amp_boot_network_list=[1, 2])
        vthunder = copy.deepcopy(VTHUNDER)
        db_task.vrrp_set_repo = mock.MagicMock()
        db_task.vrrp_set_repo.get.return_value = VRRP_SET
        setid = db_task.execute(vthunder)
        self.assertEqual(setid, a10constants.MOCK_SET_ID)

    def test_GetSpareComputeForProject(self):
        db_task = task.GetSpareComputeForProject()
        vthunder = copy.deepcopy(VTHUNDER)
        db_task.vthunder_repo = mock.MagicMock()
        db_task.vthunder_repo.get_spare_vthunder.return_value = vthunder
        db_task.vthunder_repo.set_spare_vthunder_status = mock.Mock()
        cid, vth = db_task.execute()
        self.assertEqual(cid, a10constants.MOCK_COMPUTE_ID)

    def test_TryGetSpareCompute(self):
        db_task = task.TryGetSpareCompute()
        vthunder = copy.deepcopy(VTHUNDER)
        db_task.vthunder_repo = mock.MagicMock()
        db_task.vthunder_repo.get_spare_vthunder.return_value = vthunder
        vth = db_task.execute()
        self.assertEqual(vth, vthunder)

    def test_GetComputeVThundersAndLoadBalancers(self):
        db_task = task.GetComputeVThundersAndLoadBalancers()
        vthunder = copy.deepcopy(VTHUNDER)
        db_task.vthunder_repo = mock.MagicMock()
        db_task.vthunder_repo.get_compute_vthunders.return_value = [vthunder]
        db_task.loadbalancer_repo = mock.MagicMock()
        db_task.loadbalancer_repo.get = mock.MagicMock()
        db_task.loadbalancer_repo.get.return_value = LB
        vlist, llist = db_task.execute(vthunder)
        self.assertEqual(vlist[0], vthunder)
        self.assertEqual(llist[0], LB)

    def test_DeleteStaleSpareVThunder(self):
        db_task = task.DeleteStaleSpareVThunder()
        vthunder = copy.deepcopy(VTHUNDER)
        db_task.amphora_repo = mock.Mock()
        db_task.amphora_repo.delete = mock.Mock()
        db_task.vthunder_repo = mock.Mock()
        db_task.vthunder_repo.delete = mock.Mock()
        db_task.execute(vthunder)
        db_task.amphora_repo.delete.assert_called_once_with(mock.ANY,
                                                            id=a10constants.MOCK_AMPHORA_ID)
        db_task.vthunder_repo.delete.assert_called_once_with(mock.ANY, id=vthunder.id)

    def test_GetVThunderDeviceID(self):
        db_task = task.GetVThunderDeviceID()
        vthunder = copy.deepcopy(VTHUNDER)
        db_task.amphora_repo = mock.MagicMock()
        db_task.amphora_repo.get.return_value = AMPHORA
        setid = db_task.execute(vthunder)
        self.assertEqual(setid, 1)
        db_task.amphora_repo.get.return_value = AMPHORA_BLADE
        setid = db_task.execute(vthunder)
        self.assertEqual(setid, 2)

    def test_FailoverPostDbUpdate(self):
        db_task = task.FailoverPostDbUpdate()
        vthunder = copy.deepcopy(VTHUNDER)
        db_task.amphora_repo = mock.Mock()
        db_task.amphora_repo.get.return_value = AMPHORA
        db_task.amphora_repo.update = mock.Mock()
        db_task.vthunder_repo = mock.Mock()
        db_task.vthunder_repo.get_all_vthunder_by_address.return_value = [vthunder]
        db_task.vthunder_repo.update = mock.Mock()
        db_task.execute(vthunder, vthunder)
        db_task.vthunder_repo.update.assert_called_once_with(mock.ANY,
                                                             id=vthunder.id,
                                                             ip_address=mock.ANY,
                                                             compute_id=mock.ANY)
        db_task.amphora_repo.update.assert_called_once_with(mock.ANY,
                                                            id=a10constants.MOCK_AMPHORA_ID,
                                                            compute_id=mock.ANY,
                                                            lb_network_ip=mock.ANY,
                                                            image_id=mock.ANY,
                                                            compute_flavor=mock.ANY)

    def test_count_lb_thunder_partition_success(self):
        vthunder = copy.deepcopy(VTHUNDER)
        mock_count_lb = task.CountLBThunderPartition()
        mock_count_lb.vthunder_repo.get_lb_count_vthunder_partition = mock.Mock()
        mock_count_lb.vthunder_repo.get_lb_count_vthunder_partition.return_value = 1
        lb_count = mock_count_lb.execute(vthunder)
        self.assertEqual(1, lb_count)

    def test_count_lb_thunder_partition_no_thunder(self):
        mock_count_lb = task.CountLBThunderPartition()
        lb_count = mock_count_lb.execute(None)
        self.assertEqual(0, lb_count)



#    Copyright 2020, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

SUPPORTED_FLAVOR_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "title": "Octavia Amphora Driver Flavor Metadata Schema",
    "description": "This schema is used to validate new flavor profiles "
                   "submitted for use in an a10 driver flavor profile.",
    "type": "object",
    "additionalProperties": True,
    "properties": {
        "virtual-server": {
            "type": "object",
            "description": "Specify AXAPI that will apply to the slb virtual-server",
            "properties": {
                "name-expressions": {
                    "type": "array",
                    "description": "Specify name expression to match loadbalancers "
                                   "and AXAPI that will apply to the slb virtual-server",
                    "items": {
                        "type": "object",
                        "properties": {
                            "regex": {
                                "type": "string",
                            },
                            "json": {
                                "type": "object",
                            },
                        }
                    }
                },
            }
        },
        "virtual-port": {
            "type": "object",
            "description": "Specify AXAPI that will apply to the vport",
            "properties": {
                "name-expressions": {
                    "type": "array",
                    "description": "Specify name expression to match listeners "
                                   "and AXAPI that will apply to the vport",
                    "items": {
                        "type": "object",
                        "properties": {
                            "regex": {
                                "type": "string",
                            },
                            "json": {
                                "type": "object",
                            },
                        }
                    }
                },
                "template-tcp": {
                    "type": "string",
                },
                "template-http": {
                    "type": "string",
                },
                "template-virtual-port": {
                    "type": "string",
                },
            }
        },
        "service-group": {
            "type": "object",
            "description": "Specify AXAPI that will apply to the service-group",
            "properties": {
                "name-expressions": {
                    "type": "array",
                    "description": "Specify name expression to match pools "
                                   "and AXAPI that will apply to the service-group",
                    "items": {
                        "type": "object",
                        "properties": {
                            "regex": {
                                "type": "string",
                            },
                            "json": {
                                "type": "object",
                            },
                        }
                    }
                },
            }
        },
        "server": {
            "type": "object",
            "description": "Specify AXAPI that will apply to the server",
            "properties": {
                "name-expressions": {
                    "type": "array",
                    "description": "Specify name expression to match members "
                                   "and AXAPI that will apply to the server",
                    "items": {
                        "type": "object",
                        "properties": {
                            "regex": {
                                "type": "string",
                            },
                            "json": {
                                "type": "object",
                            },
                        }
                    }
                },
                "conn-limit": {
                    "type": "integer",
                    "minimum": 1,
                    "maximum": 64000000,
                    "default": 64000000,
                },
                "conn-resume": {
                    "type": "integer",
                    "minimum": 1,
                    "maximum": 1000000,
                }
            }
        },
        "health-monitor": {
            "type": "object",
            "description": "Specify AXAPI that will apply to the health monitor",
            "properties": {
                "name-expressions": {
                    "type": "array",
                    "description": "Specify name expression to match healthmonitor "
                                   "and AXAPI that will apply to the health monitor",
                    "items": {
                        "type": "object",
                        "properties": {
                            "regex": {
                                "type": "string",
                            },
                            "json": {
                                "type": "object",
                            },
                        }
                    }
                },
            }
        },
        "nat-pool": {
            "type": "object",
            "description": "Specify AXAPI of default nat pool for loadbalancer",
            "properties": {
                "pool-name": {
                    "type": "string",
                },
                "start-address": {
                    "type": "string",
                },
                "end-address": {
                    "type": "string",
                },
                "netmask": {
                    "type": "string",
                }
            }
        },
        "nat-pool-list": {
            "type": "array",
            "description": "Specify AXAPI of nat pools for loadbalancer",
            "items": {
                "type": "object",
                "properties": {
                    "pool-name": {
                        "type": "string",
                    }
                }
            }
        },
        "device-name": {
            "type": "string",
            "description": "Specify vthunder device name that used for this loadbalancer"
        },
        "deployment": {
            "type": "object",
            "description": "Specify deployment strategy for the loadbalancer",
            "properties": {
                "dsr_type": {
                    "type": "string",
                    "description": "Specify deployment DSR type[l2dsr_transparent]"
                                   " for the loadbalancer"
                }
            }
        },
        "dns": {
            "type": "object",
            "description": "DNS nameserver information",
            "properties": {
                "primary-dns": {
                    "type": "string",
                    "description": "Primary nameserver used to contact the GLM or ELM"
                },
                "secondary-dns": {
                    "type": "string",
                    "description": "Secondary nameserver used to contact the GLM or ELM"
                }
            }
        },
        "glm-proxy-server": {
            "type": "object",
            "description": "Forward proxy-server configuration details.",
            "properties": {
                "proxy-host": {
                    "type": "string",
                    "description": "Hostname of proxy server used for requests to GLM"
                },
                "proxy-port": {
                    "type": "integer",
                    "minimum": 1,
                    "maximum": 65535,
                    "description": "Port number through which the forward proxy server "
                                   "connects to the GLM account",
                },
                "proxy-username": {
                    "type": "string",
                    "description": "Username for proxy authentication"
                },
                "proxy-password": {
                    "type": "boolean",
                    "description": "Password for proxy authentication"
                },
                "proxy-secret-string": {
                    "type": "string",
                    "description": "Password for proxy authentication"
                }
            }
        },
        "glm": {
            "type": "object",
            "description": "Global License Manager configuration settings",
            "properties": {
                "use-network-dns": {
                    "type": "boolean",
                    "description": "Use the network dns nameservers instead of "
                                   "those in the config or flavor."
                },
                "allocate-bandwidth": {
                    "type": "integer",
                    "description": "Bandwidth allocated per amphora",
                    "minimum": 2,
                    "maximum": 204800
                },
                "burst": {
                    "type": "boolean",
                    "description": "Enable bursting. Allows amphora to exceed allocated "
                                   "bandwidth limits. Ensures that packets never drop."
                },
                "interval": {
                    "type": "integer",
                    "description": "Interval for license manager heartbeat in hours",
                    "minimum": 1,
                    "maximum": 8760
                },
                "port": {
                    "type": "integer",
                    "description": "Port with which to send HTTP/S license request",
                    "minimum": 1,
                    "maximum": 65535
                },
                "enable-requests": {
                    "type": "boolean",
                    "description": "Enables license retrieval from the GLM/ELM server. "
                                   "Allows license changes to be replicated automatically.",
                },
            }
        }
    }
}



"""update_vthunders_table_with_health_state

Revision ID: 23cec8abbd48
Revises: 1722c3166462
Create Date: 2021-07-28 06:00:39.571878

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '23cec8abbd48'
down_revision = '1722c3166462'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('vthunders', sa.Column('health_state', sa.String(36), nullable=True))


def downgrade():
    op.drop_column('vthunders', 'health_state')



"""update project_id column of vrid table as owner

Revision ID: 873ee83aef63
Revises: 9bfc38df2582
Create Date: 2021-06-19 12:39:14.671476

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '873ee83aef63'
down_revision = '9bfc38df2582'
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column('vrid', 'project_id',
            new_column_name='owner',
            existing_type=sa.String(36), nullable=False)


def downgrade():
    op.alter_column('vrid', 'owner',
            new_column_name='project_id',
            existing_type=sa.String(36), nullable=False)



#    Copyright 2019, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from acos_client import errors as acos_errors
from oslo_config import cfg
from oslo_log import log as logging
from requests import exceptions
from taskflow import task

from a10_octavia.common import openstack_mappings
from a10_octavia.controller.worker.tasks.decorators import axapi_client_decorator
from a10_octavia.controller.worker.tasks.decorators import axapi_client_decorator_for_revert
from a10_octavia.controller.worker.tasks.policy import PolicyUtil
from a10_octavia.controller.worker.tasks import utils

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class L7PolicyParent(object):

    def set(self, l7policy, listeners):
        filename = l7policy.id
        p = PolicyUtil()
        script = p.createPolicy(l7policy)
        size = len(script.encode('utf-8'))
        listener = listeners[0]
        c_pers, s_pers = utils.get_sess_pers_templates(listener.default_pool)
        kargs = {}
        listener.protocol = openstack_mappings.virtual_port_protocol(self.axapi_client,
                                                                     listener.protocol)
        try:
            self.axapi_client.slb.aflex_policy.create(
                file=filename, script=script, size=size, action="import")
            LOG.debug("Successfully created l7policy: %s", l7policy.id)
        except (acos_errors.ACOSException, exceptions.ConnectionError) as e:
            LOG.exception("Failed to create/update l7policy: %s", l7policy.id)
            raise e

        try:
            get_listener = self.axapi_client.slb.virtual_server.vport.get(
                listener.load_balancer_id, listener.id,
                listener.protocol, listener.protocol_port)
            LOG.debug("Successfully fetched listener %s for l7policy %s", listener.id, l7policy.id)
        except (acos_errors.ACOSException, exceptions.ConnectionError) as e:
            LOG.exception("Failed to get listener %s for l7policy: %s", listener.id, l7policy.id)
            raise e

        if 'aflex-scripts' in get_listener['port']:
            aflex_scripts = get_listener['port']['aflex-scripts']
            aflex_scripts.append({"aflex": filename})
        else:
            aflex_scripts = [{"aflex": filename}]
        kargs["aflex_scripts"] = aflex_scripts

        try:
            self.axapi_client.slb.virtual_server.vport.update(
                listener.load_balancer_id, listener.id,
                listener.protocol, listener.protocol_port,
                listener.default_pool_id, s_pers,
                c_pers, 1, **kargs)
            LOG.debug(
                "Successfully associated l7policy %s to listener %s",
                l7policy.id,
                listener.id)
        except (acos_errors.ACOSException, exceptions.ConnectionError) as e:
            LOG.exception(
                "Failed to associate l7policy %s to listener %s",
                l7policy.id,
                listener.id)
            raise e


class CreateL7Policy(L7PolicyParent, task.Task):
    """Task to create a L7Policy"""

    @axapi_client_decorator
    def execute(self, l7policy, listeners, vthunder):
        self.set(l7policy, listeners)

    @axapi_client_decorator_for_revert
    def revert(self, l7policy, listeners, vthunder, *args, **kwargs):
        try:
            self.axapi_client.slb.aflex_policy.delete(l7policy.id)
        except exceptions.ConnectionError:
            LOG.exception("Failed to connect A10 Thunder device: %s", vthunder.ip_address)
        except Exception as e:
            LOG.warning(
                "Failed to revert creation of l7policy %s due to %s",
                l7policy.id, str(e))


class UpdateL7Policy(L7PolicyParent, task.Task):
    """Task to update L7Policy"""

    @axapi_client_decorator
    def execute(self, l7policy, listeners, vthunder, update_dict):
        l7policy.update(update_dict)
        self.set(l7policy, listeners)


class DeleteL7Policy(task.Task):
    """Task to delete L7Policy"""

    @axapi_client_decorator
    def execute(self, l7policy, vthunder):
        listener = l7policy.listener
        c_pers, s_pers = utils.get_sess_pers_templates(
            listener.default_pool)
        kargs = {}
        snat_pool = None
        if not (listener.protocol).islower():
            listener.protocol = openstack_mappings.virtual_port_protocol(
                self.axapi_client, listener.protocol)
        try:
            get_listener = self.axapi_client.slb.virtual_server.vport.get(
                listener.load_balancer_id, listener.id,
                listener.protocol, listener.protocol_port)
            if get_listener and 'port' in get_listener and 'pool' in get_listener['port']:
                snat_pool = get_listener['port']['pool']
            LOG.debug("Successfully fetched listener %s for l7policy %s", listener.id, l7policy.id)
        except (acos_errors.ACOSException, exceptions.ConnectionError) as e:
            LOG.exception("Failed to get listener %s for l7policy: %s", listener.id, l7policy.id)
            raise e

        new_aflex_scripts = []
        if 'aflex-scripts' in get_listener['port']:
            aflex_scripts = get_listener['port']['aflex-scripts']
            for aflex in aflex_scripts:
                if aflex['aflex'] != l7policy.id:
                    new_aflex_scripts.append(aflex)
        kargs["aflex_scripts"] = new_aflex_scripts

        try:
            self.axapi_client.slb.virtual_server.vport.replace(
                listener.load_balancer_id, listener.id,
                listener.protocol, listener.protocol_port,
                listener.default_pool_id,
                s_pers, c_pers, 1,
                source_nat_pool=snat_pool, **kargs)
            LOG.debug(
                "Successfully dissociated l7policy %s from listener %s",
                l7policy.id,
                listener.id)
        except (acos_errors.ACOSException, exceptions.ConnectionError) as e:
            LOG.exception(
                "Failed to dissociate l7policy %s from listener %s",
                l7policy.id,
                listener.id)
            raise e

        try:
            self.axapi_client.slb.aflex_policy.delete(l7policy.id)
            LOG.debug("Successfully deleted l7policy: %s", l7policy.id)
        except (acos_errors.ACOSException, exceptions.ConnectionError) as e:
            LOG.exception("Failed to delete l7policy: %s", l7policy.id)
            raise e



# Copyright 2020 A10 Networks
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" This module takes care of the Configuration Options set by config files for a10-octavia"""

import sys

from keystoneauth1 import loading as ks_loading
from oslo_config import cfg
from oslo_db import options as db_options
from oslo_log import log as logging
import oslo_messaging as messaging

from octavia.common import constants
from octavia.i18n import _
from octavia import version

from a10_octavia.common import a10constants
from a10_octavia.common import config_types

LOG = logging.getLogger(__name__)


A10_GLOBAL_OPTS = [
    cfg.BoolOpt('use_parent_partition', default=False,
                help=_('Use parent project partition on Thunder device '
                       'in hierarchical project architecture.')),
    cfg.IntOpt('vrid', default=0,
               help=_('VRID value')),
    cfg.StrOpt('vrid_floating_ip', default=None,
               help=_('Enable VRID floating IP feature')),
    cfg.StrOpt('network_type',
               default=a10constants.FLAT,
               choices=a10constants.SUPPORTED_NETWORK_TYPE,
               help=_('Neutron ML2 Tenent Network Type')),
    cfg.BoolOpt('use_shared_for_template_lookup',
                default=False,
                help=_('Use shared for template')),
    cfg.StrOpt('default_flavor_id', default=None,
               help=_('Default flavor ID to apply globally to all users')),
    cfg.BoolOpt('handle_vrid', default=True,
                help=_('Deletes floating ip when True.')),
]

A10_GLM_LICENSE_OPTS = [
    cfg.StrOpt('amp_license_network',
               default='',
               help=_('Network used to connect the vThunder-Amphorae '
                      'to the GLM or ELM. If not specified, the '
                      'amp_mgmt_network_list is used instead.')),
    cfg.StrOpt('flexpool_token',
               default=None,
               help=_('Alphanumeric string allowing for device licensing.')),
    cfg.StrOpt('primary_dns',
               default=None,
               help=_('Primary nameserver used to contact the GLM or ELM')),
    cfg.StrOpt('secondary_dns',
               default=None,
               help=_('Secondary nameserver used to contact the GLM or ELM')),
    cfg.IntOpt('allocate_bandwidth',
               min=2, max=204800,
               help=_('Bandwidth allocated per amphora')),
    cfg.BoolOpt('burst',
                default=False,
                help=_('Enable bursting. Allows amphora to exceed allocated '
                       'bandwidth limits. Ensures that packets never drop.')),
    cfg.IntOpt('interval',
               min=1, max=8760,
               help=_('Interval for license manager heartbeat in hours')),
    cfg.IntOpt('port',
               min=1, max=65535,
               default=443,
               help=_('Port with which to send HTTP/S license request')),
    cfg.BoolOpt('enable_requests',
                default=True,
                help=_('Enables license retrieval from the GLM/ELM server. '
                       'Allows license changes to be replicated automatically.')),
    cfg.StrOpt('proxy_host',
               default=None,
               help=_('Hostname or IP of the forward proxy')),
    cfg.StrOpt('proxy_username',
               default=None,
               help=_('Username used to authenticate to forward proxy')),
    cfg.BoolOpt('proxy_password',
                default=False,
                help=_('Password used to authenticate to forward proxy')),
    cfg.StrOpt('proxy_secret_string',
               default=None,
               help=_('Password value used to authenticate to forward proxy')),
    cfg.IntOpt('proxy_port', min=1, max=65535,
               default=None,
               help=_('Port number through which the forward proxy server '
                      'connects to the GLM account')),
]

A10_VTHUNDER_OPTS = [
    cfg.StrOpt('default_vthunder_username',
               default='admin',
               help=_('vThunder-Amphora username')),
    cfg.StrOpt('default_vthunder_password',
               default='a10',
               help=_('vThunder-Amphora password')),
    cfg.IntOpt('default_axapi_version',
               default=30,
               help=_('vThunder-Amphora AXAPI version')),
    cfg.IntOpt('default_axapi_timeout',
               default=300,
               help=_('vThunder-Amphora AXAPI timeout')),
    cfg.BoolOpt('l2dsr_support', default=False,
                help=_('For vThunder-Amphora VIP port, ingres/egress allows any address with VIP '
                       'interface MAC address to pass.')),
    cfg.BoolOpt('slb_no_snat_support', default=False,
                help=_('Allow Loadbalancer use any source address to access backend server.')),
]

A10_SLB_OPTS = [
    cfg.BoolOpt('arp_disable', default=False,
                help=_('Disable Respond to Virtual Server ARP request')),
    cfg.IntOpt('default_virtual_server_vrid',
               default=0,
               help=_('Default Virtual Server VRID')),
]

A10_HEALTH_MONITOR_OPTS = [
    cfg.StrOpt('post_data',
               help=_('HTTP Content for "--http-method POST" case.')),
]

A10_LISTENER_OPTS = [
    cfg.BoolOpt('ipinip', default=False,
                help=_('Enable IP in IP.')),
    cfg.BoolOpt('no_dest_nat',
                default=False,
                help=_('Disable destination NAT')),
    cfg.BoolOpt('ha_conn_mirror',
                default=None,
                help=_('Enable for HA Conn sync')),
    cfg.StrOpt('template_virtual_port',
               default=None,
               max_length=127,
               help=_('Provide an existing Virtual port template name on VThunder '
                      'to associate with virtual port')),
    cfg.StrOpt('template_tcp',
               default=None,
               max_length=127,
               help=_('Provide an existing TCP template name on VThunder '
                      'to associate with virtual port')),
    cfg.StrOpt('template_policy',
               default=None,
               max_length=127,
               help=_('Provide an existing Policy template name on VThunder '
                      'to associate with virtual port')),
    cfg.BoolOpt('autosnat', default=False,
                help=_('Enable autosnat')),
    cfg.IntOpt('conn_limit', min=1, max=64000000,
               default=64000000,
               help=_('Connection Limit')),
    cfg.StrOpt('template_http',
               default=None,
               max_length=127,
               help=_('Provide an existing HTTP template name on VThunder '
                      'to associate with virtual port')),
    cfg.BoolOpt('use_rcv_hop_for_resp',
                default=False,
                help=_('Use receive hop for response to client')),
]

A10_SERVICE_GROUP_OPTS = [
    cfg.StrOpt('template_server',
               default=None,
               help=_('Provide an existing Service Group Server template name on VThunder '
                      'to associate with service group')),
    cfg.StrOpt('template_port',
               default=None,
               help=_('Provide an existing Service Group Port template name on VThunder '
                      'to associate with service group')),
    cfg.StrOpt('template_policy',
               default=None,
               help=_('Provide an existing Service Group Policy template name on VThunder '
                      'to associate with service group')),
]

A10_SERVER_OPTS = [
    cfg.IntOpt('conn_limit', min=1, max=64000000,
               default=64000000,
               help=_('Connection Limit')),
    cfg.IntOpt('conn_resume', min=1, max=1000000,
               default=None,
               help=_('Connection Resume')),
    cfg.StrOpt('template_server',
               default=None,
               help=_('Provide an existing Server template name on VThunder '
                      'to associate with server')),
]

A10_HARDWARE_THUNDER_OPTS = [
    config_types.ListOfDictOpt('devices', default=[],
                               item_type=config_types.ListOfObjects(),
                               bounds=True,
                               help=_('List of all device configuration'))
]

A10_HEALTH_MANAGER_OPTS = [
    cfg.IPOpt('udp_server_ip_address',
              help=_('Server IP address that sends udp packets for '
                     'health manager.')),
    cfg.IPOpt('bind_ip', default='127.0.0.1',
              help=_('IP address the controller will listen on for '
                     'heart beats')),
    cfg.IntOpt('failover_timeout',
               default=600,
               help=_('Interval(in seconds) to wait before considering '
                      'a vThunder-Amphora is eligible for failover.')),
    cfg.IntOpt('health_check_timeout', min=1, max=180,
               default=3,
               help=_('Specify the Healthcheck timeout(in seconds) in '
                      ' vThunder-Amphora. ')),
    cfg.IntOpt('health_check_max_retries', min=1, max=10,
               default=3,
               help=_('Specify the Healthcheck Retries in '
                      'a vThunder-Amphora. ')),
    cfg.PortOpt('bind_port', default=5550,
                help=_('Port number the controller will listen on'
                       'for heart beats')),
    cfg.IntOpt('failover_threads',
               default=10,
               help=_('Number of threads performing vthunder failovers.')),
    cfg.IntOpt('status_update_threads',
               default=None,
               help=_('Number of processes for vthunder status update.')),
    cfg.IntOpt('health_update_threads',
               default=None,
               help=_('Number of processes for vthunder health update.')),
    cfg.IntOpt('stats_update_threads',
               default=4,
               help=_('Number of processes for vthunder stats update.')),
    cfg.StrOpt('heartbeat_key',
               help=_('key used to validate vthunder sending'
                      'the message'), secret=True),
    cfg.IntOpt('heartbeat_timeout',
               default=180,
               help=_('Interval(in seconds) to wait before failing over a '
                      'vthunder.')),
    cfg.IntOpt('health_check_interval',
               default=3,
               help=_('Sleep time between health checks in seconds.')),
    cfg.IntOpt('sock_rlimit', default=0,
               help=_(' sets the value of the heartbeat recv buffer')),
    cfg.ListOpt('controller_ip_port_list',
                help=_('List of controller ip and port pairs for the '
                       'heartbeat receivers. Example 127.0.0.1:5550, '
                       '192.168.0.1:5550'),
                default=[]),
    cfg.IntOpt('heartbeat_interval',
               default=10,
               help=_('Sleep time between sending heartbeats.')),
    cfg.IntOpt('stats_update_timeout',
               default=10,
               help=_('VThunder axapi timeout for Listener stats')),
    cfg.BoolOpt('stats_update_disable',
                default=False,
                help=_('Disable loadbalancer listener statistics update')),
]

A10_CONTROLLER_WORKER_OPTS = [
    cfg.IntOpt('workers',
               default=1, min=1,
               help='Number of workers for the controller-worker service.'),
    cfg.IntOpt('amp_active_retries',
               default=90,
               help=_('Retry attempts to wait for VThunder to become active')),
    cfg.IntOpt('amp_active_wait_sec',
               default=10,
               help=_('Seconds to wait between checks on whether an VThunder '
                      'has become active')),
    cfg.StrOpt('amp_flavor_id',
               default='',
               help=_('Nova instance flavor id for the VThunder')),
    cfg.StrOpt('amp_image_tag',
               default='',
               help=_('Glance image tag for the VThunder image to boot. '
                      'Use this option to be able to update the image '
                      'without reconfiguring Octavia. ')),
    cfg.StrOpt('amp_image_owner_id',
               default='',
               help=_('Restrict glance image selection to a specific '
                      'owner ID.  This is a recommended security setting.')),
    cfg.StrOpt('amp_ssh_key_name',
               default='',
               help=_('SSH key name used to boot the vThunder-Amphorae.')),
    cfg.StrOpt('amp_mgmt_network',
               default='',
               help=_('Network used to manage the vThunder-Amphorae. '
                      'If not specified, the first network in the '
                      'amp_boot_network_list is used instead.')),
    cfg.ListOpt('amp_boot_network_list',
                default='',
                help=_('List of networks to attach to the VThunder. '
                       'All networks defined in the list will '
                       'be attached to each vthunder.')),
    cfg.ListOpt('amp_secgroup_list',
                default='',
                help=_('List of security groups to attach to the vThunder-Amphorae.')),
    cfg.IntOpt('build_rate_limit',
               default=-1,
               help=_('Number of vThunder-Amphorae that could be built per controller '
                      'worker, simultaneously.')),
    cfg.StrOpt('network_driver',
               default='network_noop_driver',
               help=_('Name of the network driver to use')),
    cfg.StrOpt('loadbalancer_topology',
               default=constants.TOPOLOGY_SINGLE,
               choices=constants.SUPPORTED_LB_TOPOLOGIES,
               help=_('Load balancer topology configuration. '
                      'SINGLE - One vthunder per load balancer. '
                      'ACTIVE_STANDBY - Two vthunder per load balancer.')),
    cfg.IntOpt('amp_vcs_wait_sec',
               default=10,
               help=_('Seconds to wait between checks on whether an vThunder-Amphorae '
                      'vcs negotiation is ready')),
    cfg.IntOpt('amp_vcs_retries',
               default=60,
               help=_('Retry attempts to wait for ACOS VCS negotiation')),
    cfg.IntOpt('amp_busy_wait_sec',
               default=900,
               help=_('Timeout for waiting when vThunder-Amphora is busy. '
                      '(0 for no timeout')),
    cfg.IntOpt('retry_attempts',
               default=15,
               help=_('Retry attempts for Database Entry')),
    cfg.IntOpt('retry_initial_delay',
               default=1,
               help=_('Retry initial delay for Database Entry')),
    cfg.IntOpt('retry_bakcoff',
               default=1,
               help=_('Retry Backoff for Database Entry')),
    cfg.IntOpt('retry_max',
               default=5,
               help=_('Maximum Retries for Database Entry'))
]

A10_HOUSE_KEEPING_OPTS = [
    cfg.IntOpt('spare_check_interval',
               default=30,
               help=_('The interval in seconds for Hourkeeper to check remaining'
                      ' spare amphora in the pool')),
    cfg.IntOpt('spare_amphora_pool_size',
               default=0,
               help=_('Number of spare vThunder-Amphorae')),
    cfg.IntOpt('cleanup_interval',
               default=30,
               help=_('DB cleanup interval in seconds')),
    cfg.IntOpt('amphora_expiry_age',
               default=604800,
               help=_('vThunder-Amphora expiry age in seconds')),
    cfg.IntOpt('load_balancer_expiry_age',
               default=604800,
               help=_('Load balancer expiry age in seconds')),
    cfg.IntOpt('stats_cleanup_interval',
               default=3600,
               help=_('Statistics cleanup interval in seconds')),
    cfg.IntOpt('cert_interval',
               default=3600,
               help=_('Certificate check interval in seconds')),
    cfg.IntOpt('cert_expiry_buffer',
               default=1209600,
               help=_('Seconds until certificate expiration')),
    cfg.IntOpt('cert_rotate_threads',
               default=10,
               help=_('Number of threads performing vThunder-Amphora certificate'
                      ' rotation')),
    cfg.IntOpt('write_mem_interval',
               default=3600,
               min=300,
               help=_('Write Memory interval in seconds')),
    cfg.StrOpt('use_periodic_write_memory',
               choices=['enable', 'disable'],
               default='disable',
               help=_('Enable to use periodic write memory on all thunders'))
]

A10_NOVA_OPTS = [
    cfg.StrOpt('service_name',
               help=_('The name of the nova service in the keystone catalog')),
    cfg.StrOpt('endpoint', help=_('A new endpoint to override the endpoint '
                                  'in the keystone catalog.')),
    cfg.StrOpt('region_name',
               help=_('Region in Identity service catalog to use for '
                      'communication with the OpenStack services.')),
    cfg.StrOpt('endpoint_type', default='publicURL',
               help=_('Endpoint interface in identity service to use')),
    cfg.StrOpt('ca_certificates_file',
               help=_('CA certificates file path')),
    cfg.BoolOpt('insecure',
                default=False,
                help=_('Disable certificate validation on SSL connections')),
    cfg.BoolOpt('enable_anti_affinity', default=False,
                help=_('Flag to indicate if nova anti-affinity feature is '
                       'turned on.')),
    cfg.StrOpt('anti_affinity_policy', default=constants.ANTI_AFFINITY,
               choices=[constants.ANTI_AFFINITY, constants.SOFT_ANTI_AFFINITY],
               help=_('Sets the anti-affinity policy for nova')),
    cfg.IntOpt('random_amphora_name_length', default=0,
               help=_('If non-zero, generate a random name of the length '
                      'provided for each vthunder, in the format "a[A-Z0-9]*". '
                      'Otherwise, the default name format will be used: '
                      '"amphora-{UUID}".')),
    cfg.StrOpt('availability_zone', default=None,
               help=_('Availability zone to use for creating Amphorae')),
]


# Register the configuration options
cfg.CONF.register_opts(A10_GLOBAL_OPTS, group='a10_global')
cfg.CONF.register_opts(A10_CONTROLLER_WORKER_OPTS, group='a10_controller_worker')
cfg.CONF.register_opts(A10_HOUSE_KEEPING_OPTS, group='a10_house_keeping')
cfg.CONF.register_cli_opts(A10_HEALTH_MANAGER_OPTS, group='a10_health_manager')
cfg.CONF.register_opts(A10_NOVA_OPTS, group='a10_nova')
cfg.CONF.register_opts(A10_VTHUNDER_OPTS, group='vthunder')
cfg.CONF.register_opts(A10_GLM_LICENSE_OPTS, group='glm_license')
cfg.CONF.register_opts(A10_SLB_OPTS, group='slb')
cfg.CONF.register_opts(A10_HEALTH_MONITOR_OPTS, group='health_monitor')
cfg.CONF.register_opts(A10_LISTENER_OPTS, group='listener')
cfg.CONF.register_opts(A10_SERVICE_GROUP_OPTS, group='service_group')
cfg.CONF.register_opts(A10_SERVER_OPTS, group='server')
cfg.CONF.register_opts(A10_HARDWARE_THUNDER_OPTS, group='hardware_thunder')


# Ensure that the control exchange is set correctly
messaging.set_transport_defaults(control_exchange='octavia')
_SQL_CONNECTION_DEFAULT = 'sqlite://'
# Update the default QueuePool parameters. These can be tweaked by the
# configuration variables - max_pool_size, max_overflow and pool_timeout
db_options.set_defaults(cfg.CONF, connection=_SQL_CONNECTION_DEFAULT,
                        max_pool_size=10, max_overflow=20, pool_timeout=10)

logging.register_options(cfg.CONF)

ks_loading.register_auth_conf_options(cfg.CONF, constants.SERVICE_AUTH)
ks_loading.register_session_conf_options(cfg.CONF, constants.SERVICE_AUTH)


def init(*args, **kwargs):
    """ Initialize the cfg.CONF object for octavia project"""
    cfg.CONF(*args, project='octavia',
             version='%%prog %s' % version.version_info.release_string(),
             **kwargs)


def setup_logging(conf):
    """Sets up the logging options for a log with supplied name.

    :param conf: a cfg.ConfOpts object
    """
    product_name = "octavia"
    logging.setup(conf, product_name)
    LOG.info("Logging enabled!")
    LOG.info("%(prog)s version %(version)s",
             {'prog': sys.argv[0],
              'version': version.version_info.release_string()})
    LOG.debug("command line: %s", " ".join(sys.argv))



# Copyright (c) 2018 A10 Networks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from alembic import command as alembic_cmd
from alembic import config as alembic_cfg
from alembic import util as alembic_u
from oslo_config import cfg
from oslo_db import options
from oslo_log import log

from octavia.i18n import _

from octavia.db.migration.cli import do_alembic_command, do_check_migration, do_upgrade, no_downgrade, do_stamp, do_revision, add_command_parsers


CONF = cfg.CONF
options.set_defaults(CONF)
log.set_defaults()
log.register_options(CONF)
log.setup(CONF, 'a10-octavia-db-manage')


def main():
    config = alembic_cfg.Config(
        os.path.join(os.path.dirname(__file__), 'alembic.ini')
    )
    config.set_main_option('script_location',
                           'a10_octavia.db.migration:alembic_migrations')
    # attach the octavia conf to the Alembic conf
    config.octavia_config = CONF

    CONF(project='a10-octavia')
    CONF.command.func(config, CONF.command.name)



#    Copyright 2019, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from sqlalchemy.orm import exc as db_exceptions
import tenacity
import time
import urllib3

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
from taskflow.listeners import logging as tf_logging

from octavia.common import base_taskflow
from octavia.common import constants
from octavia.common import exceptions
from octavia.db import api as db_apis
from octavia.db import repositories as repo

from a10_octavia.common import a10constants
from a10_octavia.common import exceptions as a10_ex
from a10_octavia.common import utils
from a10_octavia.controller.worker.flows import a10_health_monitor_flows
from a10_octavia.controller.worker.flows import a10_l7policy_flows
from a10_octavia.controller.worker.flows import a10_l7rule_flows
from a10_octavia.controller.worker.flows import a10_listener_flows
from a10_octavia.controller.worker.flows import a10_load_balancer_flows
from a10_octavia.controller.worker.flows import a10_member_flows
from a10_octavia.controller.worker.flows import a10_pool_flows
from a10_octavia.controller.worker.flows import vthunder_flows
from a10_octavia.db import repositories as a10repo

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

CONF = cfg.CONF
LOG = logging.getLogger(__name__)

RETRY_ATTEMPTS = CONF.a10_controller_worker.retry_attempts
RETRY_INITIAL_DELAY = CONF.a10_controller_worker.retry_initial_delay
RETRY_BACKOFF = CONF.a10_controller_worker.retry_bakcoff
RETRY_MAX = CONF.a10_controller_worker.retry_max


def ctx_cnt_dec(ctx_lock, ctx_map, key, is_reload_thread, flags):
    LOG.debug('--------------------------ctx_cnt_dec-----------------------------')
    if flags is not None and flags[0] is False:
        return

    ctx_lock.acquire()
    try:
        ctx = ctx_map.get(key)
        if ctx is None:
            raise

        normal_thrd_num, reload_thrd_num = ctx
        LOG.debug('vthunder %s ctx: normal_thrd(%d), reload_thrd(%d)',
                  key, normal_thrd_num, reload_thrd_num)

        if is_reload_thread:
            if reload_thrd_num > 0:
                reload_thrd_num = reload_thrd_num - 1
        else:
            if normal_thrd_num > 0:
                normal_thrd_num = normal_thrd_num - 1
        if flags is not None:
            flags[0] = False
        LOG.debug('vthunder %s ctx: normal_thrd(%d), reload_thrd(%d)',
                  key, normal_thrd_num, reload_thrd_num)
        ctx_map[key] = (normal_thrd_num, reload_thrd_num)
    except Exception:
        # unexpected error should not happen, reset counters here.
        LOG.error("Unable to find vThunder instance (%s) context, reset counters.", key)
        ctx_map[key] = (0, 0)
    ctx_lock.release()


def flow_notification_handler(state, details, **kwargs):
    LOG.debug('[flow_notification_handler] state: %s', state)
    key = kwargs.get('ctx_key', None)
    if state == 'SUCCESS' or state == 'REVERTED' or state == 'FAILURE':
        is_reload_thread = kwargs.get('is_reload_thread')
        ctx_lock = kwargs.get('ctx_lock')
        ctx_map = kwargs.get('ctx_map')
        ctx_flags = kwargs.get('ctx_flags')
        if ctx_lock is None or ctx_map is None:
            raise
        ctx_cnt_dec(ctx_lock, ctx_map, key, is_reload_thread, ctx_flags)


class A10ControllerWorker(base_taskflow.BaseTaskFlowEngine):

    def __init__(self):
        self._lb_repo = repo.LoadBalancerRepository()
        self._listener_repo = repo.ListenerRepository()
        self._pool_repo = repo.PoolRepository()
        self._member_repo = a10repo.MemberRepository()
        self._health_mon_repo = repo.HealthMonitorRepository()
        self._l7policy_repo = repo.L7PolicyRepository()
        self._l7rule_repo = repo.L7RuleRepository()
        self._lb_flows = a10_load_balancer_flows.LoadBalancerFlows()
        self._listener_flows = a10_listener_flows.ListenerFlows()
        self._pool_flows = a10_pool_flows.PoolFlows()
        self._member_flows = a10_member_flows.MemberFlows()
        self._health_monitor_flows = a10_health_monitor_flows.HealthMonitorFlows()
        self._l7policy_flows = a10_l7policy_flows.L7PolicyFlows()
        self._l7rule_flows = a10_l7rule_flows.L7RuleFlows()
        self._vthunder_flows = vthunder_flows.VThunderFlows()
        self._vthunder_repo = a10repo.VThunderRepository()
        self._flavor_repo = repo.FlavorRepository()
        self._flavor_profile_repo = repo.FlavorProfileRepository()
        self._exclude_result_logging_tasks = ()
        self.ctx_map = None
        self.ctx_lock = None
        super(A10ControllerWorker, self).__init__()

    def create_amphora(self):
        """Creates an Amphora.

        This is used to create spare amphora.

        :returns: amphora_id
        """
        create_vthunder_tf = self.taskflow_load(
            self._vthunder_flows.get_create_vthunder_flow(),
            store={constants.BUILD_TYPE_PRIORITY:
                   constants.LB_CREATE_SPARES_POOL_PRIORITY,
                   constants.FLAVOR: None}
        )
        with tf_logging.DynamicLoggingListener(create_vthunder_tf, log=LOG):
            create_vthunder_tf.run()

        return create_vthunder_tf.storage.fetch('amphora')

    @tenacity.retry(
        retry=tenacity.retry_if_exception_type(db_exceptions.NoResultFound),
        wait=tenacity.wait_incrementing(
            RETRY_INITIAL_DELAY, RETRY_BACKOFF, RETRY_MAX),
        stop=tenacity.stop_after_attempt(RETRY_ATTEMPTS))
    def create_health_monitor(self, health_monitor_id):
        """Creates a health monitor.

        :param pool_id: ID of the pool to create a health monitor on
        :returns: None
        :raises NoResultFound: Unable to find the object
        """
        health_mon = self._health_mon_repo.get(db_apis.get_session(),
                                               id=health_monitor_id)
        if not health_mon:
            LOG.warning('Failed to fetch %s %s from DB. Retrying for up to '
                        '60 seconds.', 'health_monitor', health_monitor_id)
            raise db_exceptions.NoResultFound
        pool = health_mon.pool
        listeners = pool.listeners
        pool.health_monitor = health_mon
        load_balancer = pool.load_balancer

        topology = CONF.a10_controller_worker.loadbalancer_topology

        ctx_flags = [False]
        # rack flow _vthunder_busy_check() will always return False
        busy = self._vthunder_busy_check(health_mon.project_id, False, ctx_flags, load_balancer)
        try:
            create_hm_tf = self.taskflow_load(
                self._health_monitor_flows.get_create_health_monitor_flow(topology=topology),
                store={constants.HEALTH_MON: health_mon,
                       constants.POOL: pool,
                       constants.LISTENERS: listeners,
                       constants.LOADBALANCER: load_balancer,
                       a10constants.COMPUTE_BUSY: busy,
                       a10constants.WRITE_MEM_SHARED_PART: True})
            self._register_flow_notify_handler(create_hm_tf, health_mon.project_id, False,
                                               busy, ctx_flags, load_balancer)
            with tf_logging.DynamicLoggingListener(create_hm_tf,
                                                   log=LOG):
                create_hm_tf.run()
        finally:
            self._set_vthunder_available(health_mon.project_id, False, ctx_flags, load_balancer)

    def delete_health_monitor(self, health_monitor_id):
        """Deletes a health monitor.

        :param pool_id: ID of the pool to delete its health monitor
        :returns: None
        :raises HMNotFound: The referenced health monitor was not found
        """
        health_mon = self._health_mon_repo.get(db_apis.get_session(),
                                               id=health_monitor_id)

        pool = health_mon.pool
        listeners = pool.listeners
        load_balancer = pool.load_balancer

        topology = CONF.a10_controller_worker.loadbalancer_topology

        ctx_flags = [False]
        # rack flow _vthunder_busy_check() will always return False
        busy = self._vthunder_busy_check(health_mon.project_id, False, ctx_flags, load_balancer)
        try:
            delete_hm_tf = self.taskflow_load(
                self._health_monitor_flows.get_delete_health_monitor_flow(topology=topology),
                store={constants.HEALTH_MON: health_mon,
                       constants.POOL: pool,
                       constants.LISTENERS: listeners,
                       constants.LOADBALANCER: load_balancer,
                       a10constants.COMPUTE_BUSY: busy,
                       a10constants.WRITE_MEM_SHARED_PART: True})
            self._register_flow_notify_handler(delete_hm_tf, health_mon.project_id, False,
                                               busy, ctx_flags, load_balancer)
            with tf_logging.DynamicLoggingListener(delete_hm_tf,
                                                   log=LOG):
                delete_hm_tf.run()
        finally:
            self._set_vthunder_available(health_mon.project_id, False, ctx_flags, load_balancer)

    def update_health_monitor(self, health_monitor_id, health_monitor_updates):
        """Updates a health monitor.

        :param pool_id: ID of the pool to have it's health monitor updated
        :param health_monitor_updates: Dict containing updated health monitor
        :returns: None
        :raises HMNotFound: The referenced health monitor was not found
        """
        health_mon = None
        try:
            health_mon = self._get_db_obj_until_pending_update(
                self._health_mon_repo, health_monitor_id)
        except tenacity.RetryError as e:
            LOG.warning('Health monitor did not go into %s in 60 seconds. '
                        'This either due to an in-progress Octavia upgrade '
                        'or an overloaded and failing database. Assuming '
                        'an upgrade is in progress and continuing.',
                        constants.PENDING_UPDATE)
            health_mon = e.last_attempt.result()

        pool = health_mon.pool
        listeners = pool.listeners
        pool.health_monitor = health_mon
        load_balancer = pool.load_balancer

        topology = CONF.a10_controller_worker.loadbalancer_topology

        ctx_flags = [False]
        # rack flow _vthunder_busy_check() will always return False
        busy = self._vthunder_busy_check(health_mon.project_id, False, ctx_flags, load_balancer)
        try:
            update_hm_tf = self.taskflow_load(
                self._health_monitor_flows.get_update_health_monitor_flow(topology=topology),
                store={constants.HEALTH_MON: health_mon,
                       constants.POOL: pool,
                       constants.LISTENERS: listeners,
                       constants.LOADBALANCER: load_balancer,
                       a10constants.COMPUTE_BUSY: busy,
                       constants.UPDATE_DICT: health_monitor_updates,
                       a10constants.WRITE_MEM_SHARED_PART: True})
            self._register_flow_notify_handler(update_hm_tf, health_mon.project_id, False,
                                               busy, ctx_flags, load_balancer)
            with tf_logging.DynamicLoggingListener(update_hm_tf,
                                                   log=LOG):
                update_hm_tf.run()
        finally:
            self._set_vthunder_available(health_mon.project_id, False, ctx_flags, load_balancer)

    @tenacity.retry(
        retry=tenacity.retry_if_exception_type(db_exceptions.NoResultFound),
        wait=tenacity.wait_incrementing(
            RETRY_INITIAL_DELAY, RETRY_BACKOFF, RETRY_MAX),
        stop=tenacity.stop_after_attempt(RETRY_ATTEMPTS))
    def create_listener(self, listener_id):
        """Function to create listener for A10 provider"""
        listener = self._listener_repo.get(db_apis.get_session(),
                                           id=listener_id)
        if not listener:
            LOG.warning('Failed to fetch %s %s from DB. Retrying for up to '
                        '60 seconds.', 'listener', listener_id)
            raise db_exceptions.NoResultFound
        load_balancer = listener.load_balancer
        parent_project_list = utils.get_parent_project_list()
        listener_parent_proj = utils.get_parent_project(listener.project_id)

        topology = CONF.a10_controller_worker.loadbalancer_topology

        ctx_flags = [False]
        try:
            if (listener.project_id in parent_project_list or
                    (listener_parent_proj and listener_parent_proj in parent_project_list)
                    or self._is_rack_flow(listener.project_id, loadbalancer=load_balancer)):
                create_listener_tf = self.taskflow_load(
                    self._listener_flows.get_rack_vthunder_create_listener_flow(
                        listener.project_id),
                    store={constants.LOADBALANCER: load_balancer,
                           constants.LISTENER: listener})
            else:
                busy = self._vthunder_busy_check(listener.project_id, False, ctx_flags,
                                                 load_balancer)
                create_listener_tf = self.taskflow_load(
                    self._listener_flows.get_create_listener_flow(topology=topology),
                    store={constants.LOADBALANCER: load_balancer,
                           a10constants.COMPUTE_BUSY: busy,
                           constants.LISTENER: listener})
                self._register_flow_notify_handler(create_listener_tf, listener.project_id, False,
                                                   busy, ctx_flags, load_balancer)

            with tf_logging.DynamicLoggingListener(create_listener_tf,
                                                   log=LOG):
                create_listener_tf.run()
        finally:
            self._set_vthunder_available(listener.project_id, False, ctx_flags, load_balancer)

    def delete_listener(self, listener_id):
        """Function to delete a listener for A10 provider"""

        listener = self._listener_repo.get(db_apis.get_session(),
                                           id=listener_id)
        load_balancer = listener.load_balancer

        topology = CONF.a10_controller_worker.loadbalancer_topology

        ctx_flags = [False]
        try:
            if self._is_rack_flow(listener.project_id, loadbalancer=load_balancer):
                delete_listener_tf = self.taskflow_load(
                    self._listener_flows.get_delete_rack_listener_flow(),
                    store={constants.LOADBALANCER: load_balancer,
                           constants.LISTENER: listener})
            else:
                busy = self._vthunder_busy_check(listener.project_id, False, ctx_flags,
                                                 load_balancer)
                delete_listener_tf = self.taskflow_load(
                    self._listener_flows.get_delete_listener_flow(topology),
                    store={constants.LOADBALANCER: load_balancer,
                           a10constants.COMPUTE_BUSY: busy,
                           constants.LISTENER: listener})
                self._register_flow_notify_handler(delete_listener_tf, listener.project_id, False,
                                                   busy, ctx_flags, load_balancer)
            with tf_logging.DynamicLoggingListener(delete_listener_tf,
                                                   log=LOG):
                delete_listener_tf.run()
        finally:
            self._set_vthunder_available(listener.project_id, False, ctx_flags, load_balancer)

    def update_listener(self, listener_id, listener_updates):
        """Function to Update a listener for A10 provider"""

        listener = None
        try:
            listener = self._get_db_obj_until_pending_update(
                self._listener_repo, listener_id)
        except tenacity.RetryError as e:
            LOG.warning('Listener did not go into %s in 60 seconds. '
                        'This either due to an in-progress Octavia upgrade '
                        'or an overloaded and failing database. Assuming '
                        'an upgrade is in progress and continuing.',
                        constants.PENDING_UPDATE)
            listener = e.last_attempt.result()

        load_balancer = listener.load_balancer

        topology = CONF.a10_controller_worker.loadbalancer_topology

        ctx_flags = [False]
        # rack flow _vthunder_busy_check() will always return False
        busy = self._vthunder_busy_check(listener.project_id, False, ctx_flags, load_balancer)
        try:
            update_listener_tf = self.taskflow_load(
                self._listener_flows.get_update_listener_flow(topology),
                store={constants.LISTENER: listener,
                       a10constants.COMPUTE_BUSY: busy,
                       constants.LOADBALANCER: load_balancer,
                       constants.UPDATE_DICT: listener_updates})
            self._register_flow_notify_handler(update_listener_tf, listener.project_id, False,
                                               busy, ctx_flags, load_balancer)
            with tf_logging.DynamicLoggingListener(update_listener_tf, log=LOG):
                update_listener_tf.run()
        finally:
            self._set_vthunder_available(listener.project_id, False, ctx_flags, load_balancer)

    @tenacity.retry(
        retry=tenacity.retry_if_exception_type(db_exceptions.NoResultFound),
        wait=tenacity.wait_incrementing(
            RETRY_INITIAL_DELAY, RETRY_BACKOFF, RETRY_MAX),
        stop=tenacity.stop_after_attempt(RETRY_ATTEMPTS))
    def create_load_balancer(self, load_balancer_id, flavor=None, ctx_map=None, ctx_lock=None):
        """Function to create load balancer for A10 provider"""
        lb = self._lb_repo.get(db_apis.get_session(), id=load_balancer_id)
        if not lb:
            LOG.warning('Failed to fetch %s %s from DB. Retrying for up to '
                        '60 seconds.', 'load_balancer', load_balancer_id)
            raise db_exceptions.NoResultFound

        flavor_id = lb.flavor_id if lb.flavor_id else CONF.a10_global.default_flavor_id
        if not flavor and flavor_id:
            flavor = self._get_flavor_data(flavor_id)

        store = {constants.LOADBALANCER_ID: load_balancer_id,
                 constants.VIP: lb.vip,
                 constants.BUILD_TYPE_PRIORITY:
                 constants.LB_CREATE_NORMAL_PRIORITY,
                 constants.FLAVOR: flavor,
                 constants.AMPS_DATA: []}

        topology = CONF.a10_controller_worker.loadbalancer_topology

        store[constants.UPDATE_DICT] = {
            constants.TOPOLOGY: topology,
            constants.FLAVOR_ID: flavor_id
        }

        ctx_flags = [False]
        try:
            if self._is_rack_flow(lb.project_id, flavor=flavor):
                vthunder_conf = CONF.hardware_thunder.devices.get(lb.project_id, None)
                device_dict = CONF.hardware_thunder.devices
                create_lb_flow = self._lb_flows.get_create_rack_vthunder_load_balancer_flow(
                    vthunder_conf=vthunder_conf, device_dict=device_dict,
                    topology=topology, listeners=lb.listeners)
                create_lb_tf = self.taskflow_load(create_lb_flow, store=store)
            else:
                busy = self._vthunder_busy_check(lb.project_id, True, ctx_flags, lb, store)
                create_lb_flow = self._lb_flows.get_create_load_balancer_flow(
                    load_balancer_id, topology, lb.project_id, listeners=lb.listeners)
                store.update([
                    (a10constants.COMPUTE_BUSY, busy),
                    (a10constants.VTHUNDER_CONFIG, None),
                    (a10constants.USE_DEVICE_FLAVOR, False)])
                create_lb_tf = self.taskflow_load(create_lb_flow, store=store)
                self._register_flow_notify_handler(create_lb_tf, lb.project_id, True,
                                                   busy, ctx_flags, lb)

            with tf_logging.DynamicLoggingListener(
                    create_lb_tf, log=LOG,
                    hide_inputs_outputs_of=self._exclude_result_logging_tasks):
                create_lb_tf.run()
        finally:
            self._set_vthunder_available(lb.project_id, True, ctx_flags, lb)

    def delete_load_balancer(self, load_balancer_id, cascade=False):
        """Function to delete load balancer for A10 provider"""
        lb = self._lb_repo.get(db_apis.get_session(),
                               id=load_balancer_id)
        vthunder = self._vthunder_repo.get_vthunder_from_lb(db_apis.get_session(),
                                                            load_balancer_id)
        deleteCompute = False
        ctx_flags = [False]
        busy = self._vthunder_busy_check(lb.project_id, True, ctx_flags, lb)
        try:
            if vthunder:
                deleteCompute = self._vthunder_repo.get_delete_compute_flag(db_apis.get_session(),
                                                                            vthunder.compute_id)

            if cascade:
                cascade = True
            if self._is_rack_flow(lb.project_id, loadbalancer=lb):
                vthunder_conf = CONF.hardware_thunder.devices.get(lb.project_id, None)
                device_dict = CONF.hardware_thunder.devices
                (flow, store) = self._lb_flows.get_delete_rack_vthunder_load_balancer_flow(
                    lb, cascade,
                    vthunder_conf=vthunder_conf, device_dict=device_dict)
                store.update({constants.LOADBALANCER: lb,
                              a10constants.COMPUTE_BUSY: busy,
                              constants.VIP: lb.vip,
                              constants.SERVER_GROUP_ID: lb.server_group_id})
            else:
                (flow, store) = self._lb_flows.get_delete_load_balancer_flow(
                    lb, deleteCompute, cascade)
                store.update({constants.LOADBALANCER: lb,
                              a10constants.COMPUTE_BUSY: busy,
                              constants.VIP: lb.vip,
                              constants.SERVER_GROUP_ID: lb.server_group_id,
                              constants.LOADBALANCER_ID: lb.id,
                              a10constants.MASTER_AMPHORA_STATUS: True,
                              a10constants.VTHUNDER_CONFIG: None,
                              a10constants.USE_DEVICE_FLAVOR: False,
                              a10constants.LB_COUNT_THUNDER: None,
                              a10constants.MEMBER_COUNT_THUNDER: None})

            delete_lb_tf = self.taskflow_load(flow, store=store)
            self._register_flow_notify_handler(delete_lb_tf, lb.project_id, True, busy,
                                               ctx_flags, lb)

            with tf_logging.DynamicLoggingListener(delete_lb_tf,
                                                   log=LOG):
                delete_lb_tf.run()
        finally:
            self._set_vthunder_available(lb.project_id, True, ctx_flags, lb)

    def update_load_balancer(self, load_balancer_id, load_balancer_updates):
        """Function to update load balancer for A10 provider"""

        lb = None
        try:
            lb = self._get_db_obj_until_pending_update(
                self._lb_repo, load_balancer_id)
        except tenacity.RetryError as e:
            LOG.warning('Load balancer did not go into %s in 60 seconds. '
                        'This either due to an in-progress Octavia upgrade '
                        'or an overloaded and failing database. Assuming '
                        'an upgrade is in progress and continuing.',
                        constants.PENDING_UPDATE)
            lb = e.last_attempt.result()

        listeners, _ = self._listener_repo.get_all(
            db_apis.get_session(),
            load_balancer_id=load_balancer_id)

        topology = CONF.a10_controller_worker.loadbalancer_topology

        ctx_flags = [False]
        try:
            if self._is_rack_flow(lb.project_id, loadbalancer=lb):
                vthunder_conf = CONF.hardware_thunder.devices.get(lb.project_id, None)
                device_dict = CONF.hardware_thunder.devices
                update_lb_tf = self.taskflow_load(
                    self._lb_flows.get_update_rack_load_balancer_flow(vthunder_conf=vthunder_conf,
                                                                      device_dict=device_dict,
                                                                      topology=topology),
                    store={constants.LOADBALANCER: lb,
                           constants.VIP: lb.vip,
                           constants.LISTENERS: listeners,
                           constants.UPDATE_DICT: load_balancer_updates})
            else:
                busy = self._vthunder_busy_check(lb.project_id, False, ctx_flags, lb)
                update_lb_tf = self.taskflow_load(
                    self._lb_flows.get_update_load_balancer_flow(topology=topology),
                    store={constants.LOADBALANCER: lb,
                           constants.LOADBALANCER_ID: lb.id,
                           constants.VIP: lb.vip,
                           a10constants.COMPUTE_BUSY: busy,
                           constants.LISTENERS: listeners,
                           constants.UPDATE_DICT: load_balancer_updates,
                           a10constants.VTHUNDER_CONFIG: None,
                           a10constants.USE_DEVICE_FLAVOR: False})
                self._register_flow_notify_handler(update_lb_tf, lb.project_id, False,
                                                   busy, ctx_flags, lb)

            with tf_logging.DynamicLoggingListener(update_lb_tf,
                                                   log=LOG):
                update_lb_tf.run()
        finally:
            self._set_vthunder_available(lb.project_id, False, ctx_flags, lb)

    @tenacity.retry(
        retry=tenacity.retry_if_exception_type(db_exceptions.NoResultFound),
        wait=tenacity.wait_incrementing(
            RETRY_INITIAL_DELAY, RETRY_BACKOFF, RETRY_MAX),
        stop=tenacity.stop_after_attempt(RETRY_ATTEMPTS))
    def create_member(self, member_id):
        """Creates a pool member.

        :param member_id: ID of the member to create
        :returns: None
        :raises NoSuitablePool: Unable to find the node pool
        """

        member = self._member_repo.get(db_apis.get_session(),
                                       id=member_id)
        if not member:
            LOG.warning('Failed to fetch %s %s from DB. Retrying for up to '
                        '60 seconds.', 'member', member_id)
            raise db_exceptions.NoResultFound

        pool = member.pool
        listeners = pool.listeners
        load_balancer = pool.load_balancer

        topology = CONF.a10_controller_worker.loadbalancer_topology
        parent_project_list = utils.get_parent_project_list()
        member_parent_proj = utils.get_parent_project(
            member.project_id)

        ctx_flags = [False]
        try:
            if (member.project_id in parent_project_list or
                    (member_parent_proj and member_parent_proj in parent_project_list)
                    or self._is_rack_flow(member.project_id, loadbalancer=load_balancer)):
                vthunder_conf = CONF.hardware_thunder.devices.get(load_balancer.project_id, None)
                device_dict = CONF.hardware_thunder.devices
                create_member_tf = self.taskflow_load(
                    self._member_flows.get_rack_vthunder_create_member_flow(
                        vthunder_conf=vthunder_conf, device_dict=device_dict),
                    store={
                        constants.MEMBER: member,
                        constants.LISTENERS: listeners,
                        constants.LOADBALANCER: load_balancer,
                        constants.POOL: pool})
            else:
                busy = self._vthunder_busy_check(member.project_id, True, ctx_flags, load_balancer)
                create_member_tf = self.taskflow_load(
                    self._member_flows.get_create_member_flow(topology=topology),
                    store={constants.MEMBER: member,
                           constants.LISTENERS:
                           listeners,
                           constants.LOADBALANCER:
                           load_balancer,
                           a10constants.COMPUTE_BUSY: busy,
                           constants.POOL: pool,
                           a10constants.VTHUNDER_CONFIG: None,
                           a10constants.USE_DEVICE_FLAVOR: False,
                           constants.LOADBALANCER_ID: load_balancer.id})
                self._register_flow_notify_handler(create_member_tf, member.project_id, True,
                                                   busy, ctx_flags, load_balancer)

            with tf_logging.DynamicLoggingListener(create_member_tf,
                                                   log=LOG):
                create_member_tf.run()
        finally:
            self._set_vthunder_available(member.project_id, True, ctx_flags, load_balancer)

    def delete_member(self, member_id):
        """Deletes a pool member.

        :param member_id: ID of the member to delete
        :returns: None
        :raises MemberNotFound: The referenced member was not found
        """
        member = self._member_repo.get(db_apis.get_session(),
                                       id=member_id)
        pool = member.pool
        listeners = pool.listeners
        load_balancer = pool.load_balancer
        topology = CONF.a10_controller_worker.loadbalancer_topology

        ctx_flags = [False]
        try:
            if self._is_rack_flow(member.project_id, loadbalancer=load_balancer):
                vthunder_conf = CONF.hardware_thunder.devices.get(load_balancer.project_id, None)
                device_dict = CONF.hardware_thunder.devices
                delete_member_tf = self.taskflow_load(
                    self._member_flows.get_rack_vthunder_delete_member_flow(
                        vthunder_conf=vthunder_conf, device_dict=device_dict),
                    store={constants.MEMBER: member, constants.LISTENERS: listeners,
                           constants.LOADBALANCER: load_balancer, constants.POOL: pool}
                )
            else:
                busy = self._vthunder_busy_check(member.project_id, True, ctx_flags, load_balancer)
                delete_member_tf = self.taskflow_load(
                    self._member_flows.get_delete_member_flow(topology=topology),
                    store={constants.MEMBER: member, constants.LISTENERS: listeners,
                           constants.LOADBALANCER: load_balancer, a10constants.COMPUTE_BUSY: busy,
                           constants.POOL: pool, a10constants.VTHUNDER_CONFIG: None,
                           a10constants.USE_DEVICE_FLAVOR: False,
                           a10constants.LB_COUNT_THUNDER: None,
                           a10constants.MEMBER_COUNT_THUNDER: None,
                           constants.LOADBALANCER_ID: load_balancer.id}
                )
                self._register_flow_notify_handler(delete_member_tf, member.project_id, True,
                                                   busy, ctx_flags, load_balancer)
            with tf_logging.DynamicLoggingListener(delete_member_tf,
                                                   log=LOG):
                delete_member_tf.run()
        finally:
            self._set_vthunder_available(member.project_id, True, ctx_flags, load_balancer)

    def _is_batch_valid(self, old_member_ids, new_member_ids,
                        updated_member_ids, member_collision_map):
        valid = True
        for mem_id, member_col in member_collision_map.items():
            member, mem_cnt = member_col
            mem_ip = member.ip_address
            mem_port = member.protocol_port
            if mem_cnt > 1:
                if mem_id in old_member_ids:
                    error_msg = ("Duplicate members with id {} and IP {} and port {} "
                                 "found in member database.".format(mem_id, mem_ip, mem_port))
                if mem_id in new_member_ids or mem_id in updated_member_ids:
                    error_msg = ("Duplicate members with id {} and IP {} and port {} "
                                 "found in batch update request.".format(mem_id, mem_ip, mem_port))
                LOG.warning(error_msg)
                valid = False
        return valid

    def _rollback_members(self, old_member_ids, new_member_ids,
                          updated_member_ids, load_balancer,
                          listeners, pool):
        set_o_ids = set(old_member_ids)
        set_u_ids = set(updated_member_ids)
        set_n_ids = set(new_member_ids)

        current_member_ids = set_o_ids.union(set_u_ids)

        current_members = [self._member_repo.get(db_apis.get_session(), id=mid)
                           for mid in current_member_ids]

        for mem in current_members:
            # Rollback status to prevent pending state lock
            self._member_repo.update(db_apis.get_session(), mem.id,
                                     provisioning_status=constants.ACTIVE)
            LOG.info("Member with id {} and ip {} and port {} slated for "
                     "batch update have been set to ACTIVE state.".format(
                         mem.id, mem.ip_address, mem.protocol_port))

        new_members = [self._member_repo.get(db_apis.get_session(), id=mid)
                       for mid in set_n_ids]

        for mem in new_members:
            current_member_ids.add(mem.id)
            LOG.info("Member with id {} and ip {} and port {} "
                     "slated for creation under batch update "
                     "has been deleted.".format(mem.id, mem.ip_address, mem.protocol_port))
        self._member_repo.delete_members(db_apis.get_session(), set_n_ids)

        if pool is not None:
            self._pool_repo.update(db_apis.get_session(), pool.id,
                                   provisioning_status=constants.ACTIVE)
        if listeners is not None:
            for listener in listeners:
                self._listener_repo.update(db_apis.get_session(),
                                           listener.id,
                                           provisioning_status=constants.ACTIVE)
        if load_balancer is not None:
            self._lb_repo.update(db_apis.get_session(),
                                 load_balancer.id,
                                 provisioning_status=constants.ACTIVE)

    @tenacity.retry(
        retry=tenacity.retry_if_exception_type(db_exceptions.NoResultFound),
        wait=tenacity.wait_incrementing(
            RETRY_INITIAL_DELAY, RETRY_BACKOFF, RETRY_MAX),
        stop=tenacity.stop_after_attempt(RETRY_ATTEMPTS))
    def batch_update_members(self, old_member_ids, new_member_ids,
                             updated_members_req):

        updated_member_ids = [m.get('id') for m in updated_members_req]
        updated_member_models = [self._member_repo.get(db_apis.get_session(), id=mid)
                                 for mid in updated_member_ids]
        old_members = [self._member_repo.get(db_apis.get_session(), id=mid)
                       for mid in old_member_ids]
        new_members = [self._member_repo.get(db_apis.get_session(), id=mid)
                       for mid in new_member_ids]

        if old_members:
            pool = old_members[0].pool
        elif new_members:
            pool = new_members[0].pool
        else:
            pool = updated_member_models[0].pool
        listeners = pool.listeners
        load_balancer = pool.load_balancer

        modified_members = old_members + updated_member_models + new_members
        member_collision_map = {}
        for mem in modified_members:
            mem_id = mem[0] if type(mem) == tuple else mem.id
            if member_collision_map.get(mem_id):
                member_collision_map[mem_id][1] += 1
            else:
                member_collision_map[mem_id] = [mem, 1]

        if not self._is_batch_valid(old_member_ids, new_member_ids,
                                    updated_member_ids, member_collision_map):
            self._rollback_members(old_member_ids, new_member_ids,
                                   updated_member_ids, load_balancer,
                                   listeners, pool)
            LOG.warning("Due to a failed batch update caused by duplicate member definitions, "
                        "the members defined in the update are now out-of-sync with the "
                        "ACOS device. Please issue a corrected update or "
                        "delete the affected members.")
            raise a10_ex.DuplicateMembersInBatchUpdate

        # The API may not have commited all of the new member records yet.
        # Make sure we retry looking them up.
        if None in new_members or len(new_members) != len(new_member_ids):
            LOG.warning('Failed to fetch one of the new members from DB. '
                        'Retrying for up to 60 seconds.')
            raise db_exceptions.NoResultFound

        updated_members = []
        for i in range(len(updated_members_req)):
            updated_members.append((updated_member_models[i], updated_members_req[i]))

        ctx_flags = [False]
        try:
            if self._is_rack_flow(pool.project_id, loadbalancer=load_balancer):
                vthunder_conf = CONF.hardware_thunder.devices.get(load_balancer.project_id, None)
                device_dict = CONF.hardware_thunder.devices
                batch_update_members_tf = self.taskflow_load(
                    self._member_flows.get_rack_vthunder_batch_update_members_flow(
                        old_members, new_members, updated_members,
                        vthunder_conf, device_dict),
                    store={constants.LISTENERS: listeners,
                           constants.LOADBALANCER: load_balancer,
                           constants.POOL: pool})
            else:
                topology = CONF.a10_controller_worker.loadbalancer_topology
                busy = self._vthunder_busy_check(load_balancer.project_id, True,
                                                 ctx_flags, load_balancer)
                batch_update_members_tf = self.taskflow_load(
                    self._member_flows.get_batch_update_members_flow(
                        old_members, new_members, updated_members, topology),
                    store={constants.LISTENERS: listeners,
                           constants.LOADBALANCER: load_balancer,
                           a10constants.COMPUTE_BUSY: busy,
                           constants.POOL: pool,
                           constants.LOADBALANCER_ID: load_balancer.id,
                           a10constants.VTHUNDER_CONFIG: None,
                           a10constants.USE_DEVICE_FLAVOR: False,
                           a10constants.LB_COUNT_THUNDER: None,
                           a10constants.MEMBER_COUNT_THUNDER: None}
                )
                self._register_flow_notify_handler(batch_update_members_tf,
                                                   load_balancer.project_id, True,
                                                   busy, ctx_flags, load_balancer)
            with tf_logging.DynamicLoggingListener(batch_update_members_tf,
                                                   log=LOG):
                batch_update_members_tf.run()
        finally:
            self._set_vthunder_available(pool.project_id, True, ctx_flags, load_balancer)

    def update_member(self, member_id, member_updates):
        """Updates a pool member.

        :param member_id: ID of the member to update
        :param member_updates: Dict containing updated member attributes
        :returns: None
        :raises MemberNotFound: The referenced member was not found
        """
        member = None
        try:
            member = self._get_db_obj_until_pending_update(
                self._member_repo, member_id)
        except tenacity.RetryError as e:
            LOG.warning('Member did not go into %s in 60 seconds. '
                        'This either due to an in-progress Octavia upgrade '
                        'or an overloaded and failing database. Assuming '
                        'an upgrade is in progress and continuing.',
                        constants.PENDING_UPDATE)
            member = e.last_attempt.result()

        pool = member.pool
        listeners = pool.listeners
        load_balancer = pool.load_balancer

        topology = CONF.a10_controller_worker.loadbalancer_topology

        ctx_flags = [False]
        try:
            if self._is_rack_flow(member.project_id, loadbalancer=load_balancer):
                vthunder_conf = CONF.hardware_thunder.devices.get(load_balancer.project_id, None)
                device_dict = CONF.hardware_thunder.devices
                update_member_tf = self.taskflow_load(
                    self._member_flows.get_rack_vthunder_update_member_flow(
                        vthunder_conf=vthunder_conf, device_dict=device_dict),
                    store={constants.MEMBER: member,
                           constants.LISTENERS: listeners,
                           constants.LOADBALANCER: load_balancer,
                           constants.POOL: pool,
                           constants.UPDATE_DICT: member_updates})
            else:
                busy = self._vthunder_busy_check(member.project_id, False, ctx_flags,
                                                 load_balancer)
                update_member_tf = self.taskflow_load(
                    self._member_flows.get_update_member_flow(topology=topology),
                    store={constants.MEMBER: member,
                           constants.LISTENERS: listeners,
                           constants.LOADBALANCER: load_balancer,
                           constants.LOADBALANCER_ID: load_balancer.id,
                           a10constants.COMPUTE_BUSY: busy,
                           constants.POOL: pool,
                           constants.UPDATE_DICT: member_updates,
                           a10constants.VTHUNDER_CONFIG: None,
                           a10constants.USE_DEVICE_FLAVOR: False})
                self._register_flow_notify_handler(update_member_tf, member.project_id, False,
                                                   busy, ctx_flags, load_balancer)

            with tf_logging.DynamicLoggingListener(update_member_tf,
                                                   log=LOG):
                update_member_tf.run()
        finally:
            self._set_vthunder_available(member.project_id, False, ctx_flags, load_balancer)

    @tenacity.retry(
        retry=tenacity.retry_if_exception_type(db_exceptions.NoResultFound),
        wait=tenacity.wait_incrementing(
            RETRY_INITIAL_DELAY, RETRY_BACKOFF, RETRY_MAX),
        stop=tenacity.stop_after_attempt(RETRY_ATTEMPTS))
    def create_pool(self, pool_id):
        """Creates a node pool.

        :param pool_id: ID of the pool to create
        :returns: None
        :raises NoResultFound: Unable to find the object
        """
        pool = self._pool_repo.get(db_apis.get_session(),
                                   id=pool_id)
        if not pool:
            LOG.warning('Failed to fetch %s %s from DB. Retrying for up to '
                        '60 seconds.', 'pool', pool_id)
            raise db_exceptions.NoResultFound

        listeners = pool.listeners
        default_listener = None
        if listeners:
            default_listener = pool.listeners[0]
        load_balancer = pool.load_balancer

        topology = CONF.a10_controller_worker.loadbalancer_topology

        ctx_flags = [False]
        # rack flow _vthunder_busy_check() will always return False
        busy = self._vthunder_busy_check(pool.project_id, False, ctx_flags, load_balancer)
        try:
            create_pool_tf = self.taskflow_load(
                self._pool_flows.get_create_pool_flow(topology=topology),
                store={constants.POOL: pool,
                       constants.LISTENERS: listeners,
                       constants.LISTENER: default_listener,
                       constants.LOADBALANCER: load_balancer,
                       a10constants.COMPUTE_BUSY: busy})
            self._register_flow_notify_handler(create_pool_tf, pool.project_id, False,
                                               busy, ctx_flags, load_balancer)
            with tf_logging.DynamicLoggingListener(create_pool_tf,
                                                   log=LOG):
                create_pool_tf.run()
        finally:
            self._set_vthunder_available(pool.project_id, False, ctx_flags, load_balancer)

    def delete_pool(self, pool_id):
        """Deletes a node pool.

        :param pool_id: ID of the pool to delete
        :returns: None
        :raises PoolNotFound: The referenced pool was not found
        """
        pool = self._pool_repo.get(db_apis.get_session(),
                                   id=pool_id)

        load_balancer = pool.load_balancer
        listeners = pool.listeners
        default_listener = None
        if listeners:
            default_listener = pool.listeners[0]
        members = pool.members
        health_monitor = pool.health_monitor
        mem_count = self._member_repo.get_member_count(
            db_apis.get_session(),
            project_id=pool.project_id)
        mem_count = mem_count - len(members) + 1
        store = {constants.POOL: pool, constants.LISTENERS: listeners,
                 constants.LISTENER: default_listener,
                 constants.LOADBALANCER: load_balancer,
                 constants.HEALTH_MON: health_monitor,
                 a10constants.MEMBER_COUNT: mem_count,
                 constants.LOADBALANCER_ID: load_balancer.id}

        ctx_flags = [False]
        try:
            if self._is_rack_flow(pool.project_id, loadbalancer=load_balancer):
                vthunder_conf = CONF.hardware_thunder.devices.get(load_balancer.project_id, None)
                device_dict = CONF.hardware_thunder.devices
                store.update([(a10constants.VTHUNDER_CONFIG, vthunder_conf),
                              (a10constants.DEVICE_CONFIG_DICT, device_dict)])
                delete_pool_tf = self.taskflow_load(
                    self._pool_flows.get_delete_pool_rack_flow(
                        members, health_monitor, store), store=store)
            else:
                topology = CONF.a10_controller_worker.loadbalancer_topology
                store.update({a10constants.USE_DEVICE_FLAVOR: False,
                              a10constants.POOLS: None})
                busy = self._vthunder_busy_check(pool.project_id, False, ctx_flags,
                                                 load_balancer, store)
                delete_pool_tf = self.taskflow_load(
                    self._pool_flows.get_delete_pool_flow(
                        members, health_monitor, store, topology), store=store)
                self._register_flow_notify_handler(delete_pool_tf, pool.project_id, False,
                                                   busy, ctx_flags, load_balancer)

            with tf_logging.DynamicLoggingListener(delete_pool_tf,
                                                   log=LOG):
                delete_pool_tf.run()
        finally:
            self._set_vthunder_available(pool.project_id, False, ctx_flags, load_balancer)

    def update_pool(self, pool_id, pool_updates):
        """Updates a node pool.

        :param pool_id: ID of the pool to update
        :param pool_updates: Dict containing updated pool attributes
        :returns: None
        :raises PoolNotFound: The referenced pool was not found
        """

        pool = None
        try:
            pool = self._get_db_obj_until_pending_update(
                self._pool_repo, pool_id)
        except tenacity.RetryError as e:
            LOG.warning('Pool did not go into %s in 60 seconds. '
                        'This either due to an in-progress Octavia upgrade '
                        'or an overloaded and failing database. Assuming '
                        'an upgrade is in progress and continuing.',
                        constants.PENDING_UPDATE)
            pool = e.last_attempt.result()

        listeners = pool.listeners
        default_listener = None
        if listeners:
            default_listener = pool.listeners[0]
        load_balancer = pool.load_balancer

        topology = CONF.a10_controller_worker.loadbalancer_topology

        ctx_flags = [False]
        # rack flow _vthunder_busy_check() will always return False
        busy = self._vthunder_busy_check(pool.project_id, False, ctx_flags, load_balancer)
        try:
            update_pool_tf = self.taskflow_load(
                self._pool_flows.get_update_pool_flow(topology),
                store={constants.POOL: pool,
                       constants.LISTENERS: listeners,
                       constants.LISTENER: default_listener,
                       constants.LOADBALANCER: load_balancer,
                       a10constants.COMPUTE_BUSY: busy,
                       constants.UPDATE_DICT: pool_updates})
            self._register_flow_notify_handler(update_pool_tf, pool.project_id, False,
                                               busy, ctx_flags, load_balancer)
            with tf_logging.DynamicLoggingListener(update_pool_tf,
                                                   log=LOG):
                update_pool_tf.run()
        finally:
            self._set_vthunder_available(pool.project_id, False, ctx_flags, load_balancer)

    @tenacity.retry(
        retry=tenacity.retry_if_exception_type(db_exceptions.NoResultFound),
        wait=tenacity.wait_incrementing(
            RETRY_INITIAL_DELAY, RETRY_BACKOFF, RETRY_MAX),
        stop=tenacity.stop_after_attempt(RETRY_ATTEMPTS))
    def create_l7policy(self, l7policy_id):
        """Creates an L7 Policy.

        :param l7policy_id: ID of the l7policy to create
        :returns: None
        :raises NoResultFound: Unable to find the object
        """
        l7policy = self._l7policy_repo.get(db_apis.get_session(),
                                           id=l7policy_id)
        if not l7policy:
            LOG.warning('Failed to fetch %s %s from DB. Retrying for up to '
                        '60 seconds.', 'l7policy', l7policy_id)
            raise db_exceptions.NoResultFound
        listeners = [l7policy.listener]
        load_balancer = l7policy.listener.load_balancer

        topology = CONF.a10_controller_worker.loadbalancer_topology

        ctx_flags = [False]
        # rack flow _vthunder_busy_check() will always return False
        busy = self._vthunder_busy_check(l7policy.project_id, False, ctx_flags, load_balancer)
        try:
            create_l7policy_tf = self.taskflow_load(
                self._l7policy_flows.get_create_l7policy_flow(topology=topology),
                store={constants.L7POLICY: l7policy,
                       constants.LISTENERS: listeners,
                       constants.LOADBALANCER: load_balancer,
                       a10constants.COMPUTE_BUSY: busy})
            self._register_flow_notify_handler(create_l7policy_tf, l7policy.project_id, False,
                                               busy, ctx_flags, load_balancer)
            with tf_logging.DynamicLoggingListener(create_l7policy_tf,
                                                   log=LOG):
                create_l7policy_tf.run()
        finally:
            self._set_vthunder_available(l7policy.project_id, False, ctx_flags, load_balancer)

    def delete_l7policy(self, l7policy_id):
        """Deletes an L7 policy.
        :param l7policy_id: ID of the l7policy to delete
        :returns: None
        :raises L7PolicyNotFound: The referenced l7policy was not found
        """
        l7policy = self._l7policy_repo.get(db_apis.get_session(),
                                           id=l7policy_id)

        load_balancer = l7policy.listener.load_balancer
        listeners = [l7policy.listener]

        topology = CONF.a10_controller_worker.loadbalancer_topology

        ctx_flags = [False]
        # rack flow _vthunder_busy_check() will always return False
        busy = self._vthunder_busy_check(l7policy.project_id, False, ctx_flags, load_balancer)
        try:
            delete_l7policy_tf = self.taskflow_load(
                self._l7policy_flows.get_delete_l7policy_flow(topology=topology),
                store={constants.L7POLICY: l7policy,
                       constants.LISTENERS: listeners,
                       constants.LOADBALANCER: load_balancer,
                       a10constants.COMPUTE_BUSY: busy})
            self._register_flow_notify_handler(delete_l7policy_tf, l7policy.project_id, False,
                                               busy, ctx_flags, load_balancer)
            with tf_logging.DynamicLoggingListener(delete_l7policy_tf,
                                                   log=LOG):
                delete_l7policy_tf.run()
        finally:
            self._set_vthunder_available(l7policy.project_id, False, ctx_flags, load_balancer)

    def update_l7policy(self, l7policy_id, l7policy_updates):
        """Updates an L7 policy.

        :param l7policy_id: ID of the l7policy to update
        :param l7policy_updates: Dict containing updated l7policy attributes
        :returns: None
        :raises L7PolicyNotFound: The referenced l7policy was not found
        """

        l7policy = None
        try:
            l7policy = self._get_db_obj_until_pending_update(
                self._l7policy_repo, l7policy_id)
        except tenacity.RetryError as e:
            LOG.warning('L7 policy did not go into %s in 60 seconds. '
                        'This either due to an in-progress Octavia upgrade '
                        'or an overloaded and failing database. Assuming '
                        'an upgrade is in progress and continuing.',
                        constants.PENDING_UPDATE)
            l7policy = e.last_attempt.result()

        listeners = [l7policy.listener]
        load_balancer = l7policy.listener.load_balancer

        topology = CONF.a10_controller_worker.loadbalancer_topology

        ctx_flags = [False]
        # rack flow _vthunder_busy_check() will always return False
        busy = self._vthunder_busy_check(l7policy.project_id, False, ctx_flags, load_balancer)
        try:
            update_l7policy_tf = self.taskflow_load(
                self._l7policy_flows.get_update_l7policy_flow(topology=topology),
                store={constants.L7POLICY: l7policy,
                       constants.LISTENERS: listeners,
                       constants.LOADBALANCER: load_balancer,
                       a10constants.COMPUTE_BUSY: busy,
                       constants.UPDATE_DICT: l7policy_updates})
            self._register_flow_notify_handler(update_l7policy_tf, l7policy.project_id, False,
                                               busy, ctx_flags, load_balancer)
            with tf_logging.DynamicLoggingListener(update_l7policy_tf,
                                                   log=LOG):
                update_l7policy_tf.run()
        finally:
            self._set_vthunder_available(l7policy.project_id, False, ctx_flags, load_balancer)

    @tenacity.retry(
        retry=tenacity.retry_if_exception_type(db_exceptions.NoResultFound),
        wait=tenacity.wait_incrementing(
            RETRY_INITIAL_DELAY, RETRY_BACKOFF, RETRY_MAX),
        stop=tenacity.stop_after_attempt(RETRY_ATTEMPTS))
    def create_l7rule(self, l7rule_id):
        """Creates an L7 Rule.

        :param l7rule_id: ID of the l7rule to create
        :returns: None
        :raises NoResultFound: Unable to find the object
        """
        l7rule = self._l7rule_repo.get(db_apis.get_session(),
                                       id=l7rule_id)
        if not l7rule:
            LOG.warning('Failed to fetch %s %s from DB. Retrying for up to '
                        '60 seconds.', 'l7rule', l7rule_id)
            raise db_exceptions.NoResultFound
        l7policy = l7rule.l7policy
        listeners = [l7policy.listener]
        load_balancer = l7policy.listener.load_balancer

        topology = CONF.a10_controller_worker.loadbalancer_topology

        ctx_flags = [False]
        # rack flow _vthunder_busy_check() will always return False
        busy = self._vthunder_busy_check(l7rule.project_id, False, ctx_flags, load_balancer)
        try:
            create_l7rule_tf = self.taskflow_load(
                self._l7rule_flows.get_create_l7rule_flow(topology=topology),
                store={constants.L7RULE: l7rule,
                       constants.L7POLICY: l7policy,
                       constants.LISTENERS: listeners,
                       constants.LOADBALANCER: load_balancer,
                       a10constants.COMPUTE_BUSY: busy})
            self._register_flow_notify_handler(create_l7rule_tf, l7rule.project_id, False,
                                               busy, ctx_flags, load_balancer)
            with tf_logging.DynamicLoggingListener(create_l7rule_tf,
                                                   log=LOG):
                create_l7rule_tf.run()
        finally:
            self._set_vthunder_available(l7rule.project_id, False, ctx_flags, load_balancer)

    def delete_l7rule(self, l7rule_id):
        """Deletes an L7 rule.
        :param l7rule_id: ID of the l7rule to delete
        :returns: None
        :raises L7RuleNotFound: The referenced l7rule was not found
        """
        l7rule = self._l7rule_repo.get(db_apis.get_session(),
                                       id=l7rule_id)
        l7policy = l7rule.l7policy
        load_balancer = l7policy.listener.load_balancer
        listeners = [l7policy.listener]

        topology = CONF.a10_controller_worker.loadbalancer_topology

        ctx_flags = [False]
        # rack flow _vthunder_busy_check() will always return False
        busy = self._vthunder_busy_check(l7rule.project_id, False, ctx_flags, load_balancer)
        try:
            delete_l7rule_tf = self.taskflow_load(
                self._l7rule_flows.get_delete_l7rule_flow(topology=topology),
                store={constants.L7RULE: l7rule,
                       constants.L7POLICY: l7policy,
                       constants.LISTENERS: listeners,
                       constants.LOADBALANCER: load_balancer,
                       a10constants.COMPUTE_BUSY: busy})
            self._register_flow_notify_handler(delete_l7rule_tf, l7rule.project_id, False,
                                               busy, ctx_flags, load_balancer)
            with tf_logging.DynamicLoggingListener(delete_l7rule_tf,
                                                   log=LOG):
                delete_l7rule_tf.run()
        finally:
            self._set_vthunder_available(l7rule.project_id, False, ctx_flags, load_balancer)

    def update_l7rule(self, l7rule_id, l7rule_updates):
        """Updates an L7 rule.

        :param l7rule_id: ID of the l7rule to update
        :param l7rule_updates: Dict containing updated l7rule attributes
        :returns: None
        :raises L7RuleNotFound: The referenced l7rule was not found
        """
        l7rule = None
        try:
            l7rule = self._get_db_obj_until_pending_update(
                self._l7rule_repo, l7rule_id)
        except tenacity.RetryError as e:
            LOG.warning('L7 rule did not go into %s in 60 seconds. '
                        'This either due to an in-progress Octavia upgrade '
                        'or an overloaded and failing database. Assuming '
                        'an upgrade is in progress and continuing.',
                        constants.PENDING_UPDATE)
            l7rule = e.last_attempt.result()

        l7policy = l7rule.l7policy
        listeners = [l7policy.listener]
        load_balancer = l7policy.listener.load_balancer

        topology = CONF.a10_controller_worker.loadbalancer_topology

        ctx_flags = [False]
        # rack flow _vthunder_busy_check() will always return False
        busy = self._vthunder_busy_check(l7rule.project_id, False, ctx_flags, load_balancer)
        try:
            update_l7rule_tf = self.taskflow_load(
                self._l7rule_flows.get_update_l7rule_flow(topology=topology),
                store={constants.L7RULE: l7rule,
                       constants.L7POLICY: l7policy,
                       constants.LISTENERS: listeners,
                       constants.LOADBALANCER: load_balancer,
                       a10constants.COMPUTE_BUSY: busy,
                       constants.UPDATE_DICT: l7rule_updates})
            self._register_flow_notify_handler(update_l7rule_tf, l7rule.project_id, False,
                                               busy, ctx_flags, load_balancer)
            with tf_logging.DynamicLoggingListener(update_l7rule_tf,
                                                   log=LOG):
                update_l7rule_tf.run()
        finally:
            self._set_vthunder_available(l7rule.project_id, False, ctx_flags, load_balancer)

    def failover_amphora(self, vthunder_id):
        """Perform failover operations for an vThunder.
        :param vthunder_id: ID for vThunder to failover
        :returns: None
        """
        try:
            vthunder = self._vthunder_repo.get(db_apis.get_session(),
                                               vthunder_id=vthunder_id)
            if not vthunder:
                LOG.warning("Could not fetch vThunder %s from DB, ignoring "
                            "failover request.", vthunder.vthunder_id)
                return

            LOG.info("Starting Failover process on %s", vthunder.ip_address)

            store = {a10constants.VTHUNDER: vthunder,
                     a10constants.FAILOVER_VTHUNDER: vthunder}
            failover_tf = None
            if vthunder.topology == a10constants.TOPOLOGY_SPARE:
                failover_tf = self.taskflow_load(
                    self._vthunder_flows.get_failover_spare_vthunder_flow(),
                    store=store)
            elif vthunder.topology == constants.TOPOLOGY_ACTIVE_STANDBY:
                health_vthunder_count = self._vthunder_repo.get_health_vthunder_count_for_lb(
                    db_apis.get_session(), vthunder.loadbalancer_id)
                if health_vthunder_count > 0:
                    failover_tf = self.taskflow_load(
                        self._vthunder_flows.get_failover_vcs_vthunder_flow(),
                        store=store)
                else:
                    LOG.warning("Failover for a total HA Pair failure is not supported. "
                                "Pair will be kept in failed state.")
                    failover_tf = self.taskflow_load(
                        self._vthunder_flows.get_failover_restore_vthunder_flow(),
                        store=store)

            if failover_tf:
                with tf_logging.DynamicLoggingListener(failover_tf, log=LOG):
                    failover_tf.run()

        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.error("vThunder %(id)s topology %(topology)s failover exception: %(exc)s",
                          {'id': vthunder_id, 'topology': vthunder.topology, 'exc': e})

    def failover_loadbalancer(self, load_balancer_id):
        """Perform failover operations for a load balancer.

        :param load_balancer_id: ID for load balancer to failover
        :returns: None
        :raises LBNotFound: The referenced load balancer was not found
        """

        raise exceptions.NotImplementedError(
            user_fault_string='This provider does not support loadbalancer '
                              'failover yet.',
            operator_fault_string='This provider does not support loadbalancer '
                                  'failover yet.')

    def amphora_cert_rotation(self, amphora_id):
        """Perform cert rotation for an amphora.

        :param amphora_id: ID for amphora to rotate
        :returns: None
        :raises AmphoraNotFound: The referenced amphora was not found
        """

        raise exceptions.NotImplementedError(
            user_fault_string='This provider does not support rotating Amphora '
                              'certs.',
            operator_fault_string='This provider does not support rotating '
                                  'Amphora certs. We will use preconfigured '
                                  'devices.')

    def _get_db_obj_until_pending_update(self, repo, id):
        return repo.get(db_apis.get_session(), id=id)

    def perform_write_memory(self, thunders):
        """Perform write memory operations for a thunders

        :param thunders: group of thunder objects
        :returns: None
        """
        store = {a10constants.WRITE_MEM_SHARED_PART: True}

        for vthunder in thunders:
            delete_compute = False
            if vthunder.status == 'DELETED' and vthunder.compute_id is not None:
                delete_compute = self._vthunder_repo.get_delete_compute_flag(db_apis.get_session(),
                                                                             vthunder.compute_id)
            try:
                write_mem_tf = self.taskflow_load(
                    self._vthunder_flows.get_write_memory_flow(vthunder, store, delete_compute),
                    store=store)

                with tf_logging.DynamicLoggingListener(write_mem_tf,
                                                       log=LOG):
                    write_mem_tf.run()
            except Exception:
                # continue on other thunders (assume exception is logged)
                pass

    def perform_reload_check(self, thunders):
        """Perform check for thunders see if thunder reload before write memory

        :param thunders: group of thunder objects
        :returns: None
        """
        store = {}
        for vthunder in thunders:
            try:
                reload_check_tf = self.taskflow_load(
                    self._vthunder_flows.get_reload_check_flow(vthunder, store),
                    store=store)
                with tf_logging.DynamicLoggingListener(reload_check_tf, log=LOG):
                    reload_check_tf.run()
            except Exception:
                # continue on other thunders (assume exception is logged)
                pass

    def perform_vthunder_stats_update(self, ip):
        """Perform for listener statistics update"""

        store = {}
        try:
            thunders = self._vthunder_repo.get_all_vthunder_by_address(
                db_apis.get_session(),
                ip_address=ip)
            for vthunder in thunders:
                vthunder_stats_tf = self.taskflow_load(
                    self._listener_flows.get_listener_stats_flow(vthunder, store),
                    store=store)
                with tf_logging.DynamicLoggingListener(vthunder_stats_tf, log=LOG):
                    vthunder_stats_tf.run()
        except Exception:
            # continue on other thunders (assume exception is logged)
            pass

    def a10_worker_ctx_init(self, ctx_map, ctx_lock):
        self.ctx_map = ctx_map
        self.ctx_lock = ctx_lock

    def _is_rack_flow(self, key, loadbalancer=None, flavor=None):
        if self.ctx_map is None or self.ctx_lock is None:
            return True
        if key in CONF.hardware_thunder.devices:
            return True

        # rack flow with device flavor
        device_name = None
        if flavor:
            device_name = flavor.get('device-name', None)
        elif loadbalancer:
            flavor_data = utils.get_loadbalancer_flavor(loadbalancer)
            if flavor_data is not None:
                device_name = flavor_data.get('device-name', None)
        if device_name is not None and isinstance(CONF.hardware_thunder.devices, dict):
            for device in CONF.hardware_thunder.devices.values():
                if device.device_name == device_name:
                    return True

        return False

    def _vthunder_busy_check(self, key, is_reload_thread, flags, loadbalancer, store=None):
        busy = False
        if self._is_rack_flow(key, loadbalancer=loadbalancer):
            return busy

        timeout = CONF.a10_controller_worker.amp_busy_wait_sec
        while timeout >= 0:
            # amp_busy_wait_sec 0 for wait forever
            if CONF.a10_controller_worker.amp_busy_wait_sec != 0:
                timeout = timeout - 5

            self.ctx_lock.acquire()
            ctx = self.ctx_map.get(key, None)
            if ctx is None:
                ctx = (0, 0)
            normal_thrd_num, reload_thrd_num = ctx
            LOG.debug('[busy_check] vthunder %s ctx: normal_thrd(%d), reload_thrd(%d)',
                      key, normal_thrd_num, reload_thrd_num)
            if is_reload_thread:
                if reload_thrd_num > 0 or normal_thrd_num > 0:
                    busy = True
                else:
                    reload_thrd_num = reload_thrd_num + 1
                    self.ctx_map[key] = (normal_thrd_num, reload_thrd_num)
                    busy = False
            else:
                if reload_thrd_num > 0:
                    busy = True
                else:
                    normal_thrd_num = normal_thrd_num + 1
                    self.ctx_map[key] = (normal_thrd_num, reload_thrd_num)
                    busy = False
            self.ctx_lock.release()

            if not busy:
                LOG.debug('[busy_check] vthunder %s ctx: normal_thrd(%d), reload_thrd(%d)',
                          key, normal_thrd_num, reload_thrd_num)
                flags[0] = True
                break
            time.sleep(5)

        if store is not None:
            store[a10constants.COMPUTE_BUSY] = busy

        return busy

    def _set_vthunder_available(self, key, is_reload_thread, flags, loadbalancer):
        if self._is_rack_flow(key, loadbalancer=loadbalancer):
            return
        ctx_cnt_dec(self.ctx_lock, self.ctx_map, key, is_reload_thread, flags)

    def _register_flow_notify_handler(self, engine, key, is_reload_thread, instance_busy,
                                      flags, loadbalancer):
        if self._is_rack_flow(key, loadbalancer=loadbalancer) or instance_busy:
            return
        kwargs = {'ctx_key': key, 'ctx_lock': self.ctx_lock, 'ctx_map': self.ctx_map,
                  'is_reload_thread': is_reload_thread, 'ctx_flags': flags}
        engine.notifier.register('*', flow_notification_handler, kwargs=kwargs)

    def _get_flavor_data(self, flavor_id):
        flavor = self._flavor_repo.get(db_apis.get_session(), id=flavor_id)
        if flavor and flavor.flavor_profile_id:
            flavor_profile = self._flavor_profile_repo.get(
                db_apis.get_session(),
                id=flavor.flavor_profile_id)
            flavor_data = json.loads(flavor_profile.flavor_data)
            return flavor_data
        return None



"""Updated vrid table to use a uuid for its primary key

Revision ID: 9bfc38df2582
Revises: cfce01f4010d
Create Date: 2021-05-17 08:42:42.132368

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import mysql

# revision identifiers, used by Alembic.
revision = '9bfc38df2582'
down_revision = 'cfce01f4010d'
branch_labels = None
depends_on = None


def upgrade():
    op.execute('ALTER TABLE vrid DROP PRIMARY KEY, CHANGE id id int(11)')
    op.alter_column('vrid', 'id',
        existing_type=mysql.VARCHAR(length=36),
        type_=sa.String(36))
    op.execute('ALTER TABLE vrid ADD PRIMARY KEY (id)')
    op.alter_column('vrid', 'vrid_floating_ip',
        existing_type=mysql.VARCHAR(length=40),
        nullable=False)

def downgrade():
    op.execute('ALTER TABLE vrid DROP PRIMARY KEY')
    op.alter_column('vrid', 'id', type_=sa.Integer)
    op.execute('ALTER TABLE vrid ADD PRIMARY KEY (id)')
    op.execute('ALTER TABLE vrid MODIFY id INTEGER NOT NULL AUTO_INCREMENT')
    op.alter_column('vrid', 'vrid_floating_ip',
        existing_type=mysql.VARCHAR(length=40),
        nullable=True)



#    Copyright 2020, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import acos_client.errors as acos_errors
import copy
import imp
try:
    from unittest import mock
except ImportError:
    import mock

from octavia.common import data_models as o_data_models

from a10_octavia.common.data_models import VThunder
from a10_octavia.controller.worker.tasks import persist_tasks
from a10_octavia.tests.common import a10constants as a10_test_constants
from a10_octavia.tests.unit.base import BaseTaskTestCase

VTHUNDER = VThunder()
POOL = o_data_models.Pool(id=a10_test_constants.MOCK_POOL_ID)
SOURCE_IP_SESS_PERS = o_data_models.SessionPersistence(
    pool_id=a10_test_constants.MOCK_POOL_ID,
    type='SOURCE_IP')
APP_COOKIE_SESS_PERS = o_data_models.SessionPersistence(
    pool_id=a10_test_constants.MOCK_POOL_ID,
    type='APP_COOKIE', cookie_name=a10_test_constants.MOCK_COOKIE_NAME)
HTTP_COOKIE_SESS_PERS = o_data_models.SessionPersistence(
    pool_id=a10_test_constants.MOCK_POOL_ID,
    type='HTTP_COOKIE', cookie_name=a10_test_constants.MOCK_COOKIE_NAME)
APP_COOKIE_SESS_PERS_WITH_NO_TYPE = o_data_models.SessionPersistence(
    pool_id=a10_test_constants.MOCK_POOL_ID,
    cookie_name=a10_test_constants.MOCK_COOKIE_NAME)


class TestPersistTasks(BaseTaskTestCase):

    def setUp(self):
        super(TestPersistTasks, self).setUp()
        imp.reload(persist_tasks)
        self.pool = copy.deepcopy(POOL)
        self.client_mock = mock.Mock()

    def test_handle_session_persistence_with_source_ip(self):
        mock_session_persist = persist_tasks.HandleSessionPersistenceDelta()
        mock_session_persist.axapi_client = self.client_mock
        self.pool.session_persistence = SOURCE_IP_SESS_PERS
        mock_session_persist.execute(VTHUNDER, self.pool)
        self.client_mock.slb.template.src_ip_persistence.delete.assert_called_with(
            POOL.id)
        self.client_mock.slb.template.src_ip_persistence.create.assert_called_with(
            POOL.id)

    def test_handle_session_persistence_with_app_cookie(self):
        mock_session_persist = persist_tasks.HandleSessionPersistenceDelta()
        mock_session_persist.axapi_client = self.client_mock
        self.pool.session_persistence = APP_COOKIE_SESS_PERS
        mock_session_persist.execute(VTHUNDER, self.pool)
        self.client_mock.slb.template.cookie_persistence.delete.assert_called_with(
            POOL.id)
        self.client_mock.slb.template.cookie_persistence.create.assert_called_with(
            POOL.id, cookie_name=a10_test_constants.MOCK_COOKIE_NAME)

    def test_handle_session_persistence_with_http_cookie(self):
        mock_session_persist = persist_tasks.HandleSessionPersistenceDelta()
        mock_session_persist.axapi_client = self.client_mock
        self.pool.session_persistence = HTTP_COOKIE_SESS_PERS
        mock_session_persist.execute(VTHUNDER, self.pool)
        self.client_mock.slb.template.cookie_persistence.delete.assert_called_with(
            POOL.id)
        self.client_mock.slb.template.cookie_persistence.create.assert_called_with(
            POOL.id, cookie_name=a10_test_constants.MOCK_COOKIE_NAME)

    def test_handle_session_persistence_with_app_cookie_with_no_sess_pers_type(self):
        mock_session_persist = persist_tasks.HandleSessionPersistenceDelta()
        mock_session_persist.axapi_client = self.client_mock
        self.pool.session_persistence = APP_COOKIE_SESS_PERS_WITH_NO_TYPE
        mock_session_persist.execute(VTHUNDER, self.pool)
        self.client_mock.slb.template.cookie_persistence.delete.assert_called_with(POOL.id)
        self.client_mock.slb.template.cookie_persistence.create.assert_not_called()

    def test_handle_session_persistence_with_source_ip_with_no_sess_pers(self):
        mock_session_persist = persist_tasks.HandleSessionPersistenceDelta()
        mock_session_persist.axapi_client = self.client_mock
        mock_session_persist.execute(VTHUNDER, self.pool)
        self.client_mock.slb.template.src_ip_persistence.delete.assert_called_with(POOL.id)
        self.client_mock.slb.template.src_ip_persistence.create.assert_not_called()

    def test_revert_handle_session_persistence_with_source_ip(self):
        mock_session_persist = persist_tasks.HandleSessionPersistenceDelta()
        mock_session_persist.axapi_client = self.client_mock
        self.pool.session_persistence = SOURCE_IP_SESS_PERS
        mock_session_persist.execute(VTHUNDER, self.pool)
        self.client_mock.slb.template.src_ip_persistence.delete.assert_called_with(
            POOL.id)

    def test_revert_handle_session_persistence_with_http_cookie(self):
        mock_session_persist = persist_tasks.HandleSessionPersistenceDelta()
        mock_session_persist.axapi_client = self.client_mock
        self.pool.session_persistence = HTTP_COOKIE_SESS_PERS
        mock_session_persist.execute(VTHUNDER, self.pool)
        self.client_mock.slb.template.cookie_persistence.delete.assert_called_with(
            POOL.id)

    def test_handle_session_persistence_exception(self):
        mock_session_persist = persist_tasks.HandleSessionPersistenceDelta()
        mock_session_persist.axapi_client = self.client_mock
        self.client_mock.side_effect = acos_errors.Exists
        mock_session_persist.execute(VTHUNDER, self.pool)

    def test_delete_session_persistence(self):
        mock_session_persist = persist_tasks.DeleteSessionPersistence()
        mock_session_persist.axapi_client = self.client_mock
        self.pool.session_persistence = APP_COOKIE_SESS_PERS
        mock_session_persist.execute(VTHUNDER, self.pool)
        self.client_mock.slb.template.cookie_persistence.delete.assert_called_with(POOL.id)

    def test_delete_session_persistence_with_no_sess_pers(self):
        mock_session_persist = persist_tasks.DeleteSessionPersistence()
        mock_session_persist.axapi_client = self.client_mock
        mock_session_persist.execute(VTHUNDER, self.pool)
        self.client_mock.slb.template.cookie_persistence.delete.assert_not_called()

    def test_unset_session_persistence(self):
        mock_session_persist = persist_tasks.HandleSessionPersistenceDelta()
        mock_session_persist.axapi_client = self.client_mock
        self.pool.session_persistence = None
        mock_session_persist.execute(VTHUNDER, self.pool)
        self.client_mock.slb.template.src_ip_persistence.delete.assert_called_with(POOL.id)
        self.client_mock.slb.template.src_ip_persistence.create.assert_not_called()
